# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial.transform import Rotation as R


def pretty_plot():
    """
    Makes pretty plots. Overwrites the matplotlib default settings to allow
    for better fonts. Slows down plotting
    """
    import matplotlib.pyplot as plt

    plt.rc("text", usetex=True)
    plt.rcParams["text.latex.preamble"] = "\\usepackage{relsize}"
    plt.rc("font", family="sanserif")


def _to_polar(vecs):
    """
    Convenience function converting x, y, z vectors to
    normalised positions on a polar plot.

    :param vecs: Array of vectors to convert
    :type vecs: numpy array
    :return: Arrays of azimuths and distances from the origin
    :rtype: Two numpy arrays
    """
    shape = vecs.shape
    vecs = vecs.reshape(shape[0] * shape[1], 3)
    Rs = np.linalg.norm(vecs, axis=-1)

    zeniths = np.arccos(-vecs[:, 2] / Rs)
    thetas = (np.arctan2(vecs[:, 1], vecs[:, 0])).reshape(shape[0], shape[1])
    r = (np.sin(zeniths) / (1.0 - np.cos(zeniths))).reshape(shape[0], shape[1])
    return thetas, r


def _positions(
    n_zenith,
    n_azimuth,
    min_zenith,
    max_zenith,
    latitude_direction=False,
    azimuth_midpoints=False,
):
    """
    Convenience function creating a set of equally spaced
    x, y, z vectors and theta and r values on a polar plot.

    :param n_zenith: Number of zeniths
    :type n_zenith: integer
    :param n_azimuth: Number of azimuths
    :type n_azimuth: integer
    :param min_zenith: _description_
    :type min_zenith: float
    :param max_zenith: _description_
    :type max_zenith: float
    :param latitude_direction: Whether to return both positions and vectors
        along local lines of latitude,
        defaults to False
    :type latitude_direction: bool, optional
    :param azimuth_midpoints: Whether to use azimuthal midpoints,
        rather than starting at 0 and ending at 2*pi. defaults to False
    :type azimuth_midpoints: bool, optional
    :return: Either a full basis or just the positional x, y, z vectors,
        azimuth values and r values for plotting on a polar plot.
    :rtype: Three or five numpy arrays
    """
    # Upper hemisphere plot, zenith starts at pi/2, increases to pi.
    zeniths = np.linspace(min_zenith, max_zenith, n_zenith)
    if azimuth_midpoints:
        azimuths = np.linspace(
            (np.pi / (n_azimuth + 1)),
            2.0 * np.pi - (np.pi / (n_azimuth + 1)),
            n_azimuth,
        )
    else:
        azimuths = np.linspace(0.0, 2.0 * np.pi, n_azimuth)

    Rs = np.sin(zeniths) / (1.0 - np.cos(zeniths))
    r, theta = np.meshgrid(Rs, azimuths)

    cos_az = np.cos(azimuths)
    sin_phi = np.sin(zeniths)
    sin_az = np.sin(azimuths)
    cos_phi = np.cos(zeniths)

    cos_phi, cos_az = np.meshgrid(cos_phi, cos_az)
    sin_phi, sin_az = np.meshgrid(sin_phi, sin_az)

    d = np.moveaxis(np.array([cos_az * sin_phi, sin_az * sin_phi, -cos_phi]), 0, 2)

    if latitude_direction:
        d1 = np.moveaxis(np.array([cos_az * cos_phi, sin_az * cos_phi, sin_phi]), 0, 2)
        return d, d1, theta, r
    else:
        return d, theta, r


def _rotated_vector_sets(rotation_axes, vector_initial, n_rots):
    """
    Convenience function to calculate a set of
    vectors rotated between 0 and 180 degrees
    from their initial positions about an axis.

    :param rotation_axes: rotation axes
    :type rotation_axes: numpy array
    :param vector_1_initial: vectors to be rotated
    :type vector_1_initial: numpy array
    :param vector_2_initial: another set of vectors to be rotated
    :type vector_2_initial: numpy array
    :param n_rots: number of rotations between 0 and 180 degrees.
    :type n_rots: integer
    :return: A new set of vectors
    :rtype: Two numpy arrays
    """
    # now rotate d1 and d2 about d in 1 degree increments
    n_azimuth = vector_initial.shape[0]
    n_zenith = vector_initial.shape[1]
    rotations = np.linspace(0.0, np.pi, n_rots)
    axial_directions = np.empty((n_azimuth, n_zenith, n_rots, 3))
    lateral_directions = np.empty((n_azimuth, n_zenith, n_rots, 3))
    for i in range(n_rots):
        axial_directions[:, :, i, :] = rotation_axes
    for i in range(n_azimuth):
        for j in range(n_zenith):
            rotvecs = np.einsum("i, l", rotations, rotation_axes[i, j])
            rot = R.from_rotvec(rotvecs)
            lateral_directions[i, j] = rot.apply(vector_initial[i, j])
    return axial_directions, lateral_directions


def _make_lines(positions, vectors, scales):
    """
    Convenience function to make line segments for polar plots.

    :param positions: Center position (x, y, z)
    :type positions: numpy array
    :param vectors: Vector away from the center position (also Cartesian)
    :type vectors: numpy array
    :param scales: Scaling fractions for the length of line.
    :type scales: numpy array
    :return: The requested line segments
    :rtype: numpy array
    """
    vec_start = positions - np.einsum("ijk, ij->ijk", vectors, scales)
    vec_end = positions + np.einsum("ijk, ij->ijk", vectors, scales)
    n_lines = positions.shape[0] * positions.shape[1]

    vlines = np.array([_to_polar(vec_start), _to_polar(vec_end)]).reshape(2, 2, n_lines)
    vlines = np.moveaxis(vlines, (0, 2), (2, 0))
    return vlines


def _make_poisson_lines(mineral, n_rots, vector_fraction, plot_minimum=False):
    """
    Convenience function to make line segments
    for the polar plot to indicate the orientations of
    the minimum or maximum Poisson ratio eigenvectors

    :param mineral: Anisotropic mineral to plot.
    :type mineral: AnisotropicMineral
    :param n_rots: Number of rotations to find the minimum
    :type n_rots: integer
    :param vector_fraction: Fractional length of plotted unit vector.
    :type vector_fraction: float
    :param plot_minimum: Whether to plot the minimum (or maximum),
        defaults to False
    :type plot_minimum: bool, optional

    :return: The requested line segments.
    :rtype: numpy array
    """
    v = []
    for i in range(3):
        n_zenith = 2
        n_azimuth = 36 - 16 * i  # 36 - 16*2 = 4
        min_zenith = (7.0 + 2 * i) * np.pi / 13.0
        max_zenith = (8.0 + 2 * i) * np.pi / 13.0
        d0, d01, _, r = _positions(
            n_zenith, n_azimuth, min_zenith, max_zenith, True, True
        )

        axial, lateral = _rotated_vector_sets(d0, d01, n_rots)
        poisson0 = mineral.isentropic_poissons_ratio(axial, lateral)

        if plot_minimum:
            inds0 = np.ix_(*[np.arange(i) for i in poisson0.shape[:-1]]) + (
                np.argmin(poisson0, axis=-1),
            )
            v.extend(
                list(_make_lines(d0, lateral[inds0], vector_fraction / np.sqrt(r)))
            )
        else:
            inds0 = np.ix_(*[np.arange(i) for i in poisson0.shape[:-1]]) + (
                np.argmax(poisson0, axis=-1),
            )
            v.extend(
                list(_make_lines(d0, lateral[inds0], vector_fraction / np.sqrt(r)))
            )
    return np.array(v)


def _make_velocity_lines(mineral, vector_fraction):
    """
    Convenience function to make line segments
    for the polar plot to indicate the orientations of
    the VS1 and VS2 eigenvectors

    :param mineral: Anisotropic mineral to plot.
    :type mineral: AnisotropicMineral
    :param vector_fraction: Fractional length of plotted unit vector.
    :type vector_fraction: float

    :return: The requested line segments.
    :rtype: Two 2D numpy arrays
    """
    v_vs1 = []
    v_vs2 = []
    for i in range(3):
        n_zenith = 2
        n_azimuth = 36 - 16 * i  # 36 - 16*2 = 4
        min_zenith = (7.0 + 2 * i) * np.pi / 13.0
        max_zenith = (8.0 + 2 * i) * np.pi / 13.0
        d0, _, r = _positions(n_zenith, n_azimuth, min_zenith, max_zenith, False, True)

        _, eigenvectors = mineral.wave_velocities(d0)

        v_vs1.extend(
            list(_make_lines(d0, eigenvectors[:, :, 1], vector_fraction / np.sqrt(r)))
        )
        v_vs2.extend(
            list(_make_lines(d0, eigenvectors[:, :, 2], vector_fraction / np.sqrt(r)))
        )
    return np.array(v_vs1), np.array(v_vs2)


def plot_projected_elastic_properties(
    mineral,
    plot_types,
    axes,
    n_zenith=31,
    n_azimuth=91,
    n_divs=100,
    n_rots=181,
    plot_vectors=True,
):
    """
    :param mineral: Mineral object on which calculations should be done
    :type mineral: :class:`burnman.Mineral`

    :param plot_types: Plot types must be one of the following
        * 'vp' - V$_{P}$ (km/s)
        * 'vs1' - 'V$_{S1}$ (km/s)
        * 'vs2' - V$_{S2}$ (km/s)
        * 'vp/vs1' - V$_{P}$/V$_{S1}$
        * 'vp/vs2' - V$_{P}$/V$_{S2}$
        * 's anisotropy' - S-wave anisotropy (%), 200(vs1s - vs2s)/(vs1s + vs2s))
        * 'linear compressibility' - Linear compressibility (GPa$^{-1}$)
        * 'youngs modulus' - Youngs Modulus (GPa)
        * 'minimum poisson ratio' - Minimum Poisson ratio
        * 'maximum poisson ratio' - Maximum Poisson ratio
    :type plot_types: list of str

    :param axes: axes objects to be modified.
        Must be initialised with projection='polar'.
    :type axes: matplotlib.pyplot.axes objects

    :param n_zenith: Number of zeniths (plot resolution).
    :type n_zenith: int

    :param n_azimuth: Number of azimuths (plot resolution).
    :type n_azimuth: int

    :param n_divs: Number of divisions for the color scale.
    :type n_divs: int

    :param n_rots: Number of along-vector rotations to
        find minimum and maximum Poisson ratios.
    :type n_rots: int

    :param plot_vectors: Whether or not to plot vectors
        for $V_S$ and Poisson axial directions
    :type plot_vectors: bool
    """

    assert len(plot_types) == len(axes)

    # Upper hemisphere plot, zenith starts at pi/2, increases to pi.
    d, d1, theta, r = _positions(n_zenith, n_azimuth, np.pi / 2.0, np.pi, True)

    velocities, _ = mineral.wave_velocities(d)
    vps, vs1s, vs2s = np.moveaxis(velocities, -1, 0)
    v_vs1s, v_vs2s = _make_velocity_lines(mineral, 0.02)

    if "minimum poisson ratio" in plot_types or "maximum poisson ratio" in plot_types:
        axial_directions, lateral_directions = _rotated_vector_sets(d, d1, n_rots)
        poisson = mineral.isentropic_poissons_ratio(
            axial_directions, lateral_directions
        )

    prps = []
    for type in plot_types:
        if type == "vp":
            prps.append(("V$_{P}$ (km/s)", vps / 1000.0))
        elif type == "vs1":
            prps.append(("V$_{S1}$ (km/s)", (vs1s / 1000.0, v_vs1s)))
        elif type == "vs2":
            prps.append(("V$_{S2}$ (km/s)", (vs2s / 1000.0, v_vs2s)))
        elif type == "vp/vs1":
            prps.append(("V$_{P}$/V$_{S1}$", vps / vs1s))
        elif type == "vp/vs2":
            prps.append(("V$_{P}$/V$_{S2}$", vps / vs2s))
        elif type == "s anisotropy":
            prps.append(
                ("S-wave anisotropy (%)", 200.0 * (vs1s - vs2s) / (vs1s + vs2s))
            )
        elif type == "linear compressibility":
            betas = mineral.isentropic_linear_compressibility(d)
            prps.append(("Linear compressibility (GPa$^{-1}$)", betas * 1.0e9))
        elif type == "youngs modulus":
            Es = mineral.isentropic_youngs_modulus(d)
            prps.append(("Youngs Modulus (GPa)", Es / 1.0e9))
        elif type == "minimum poisson ratio":
            inds = np.ix_(*[np.arange(i) for i in poisson.shape[:-1]]) + (
                np.argmin(poisson, axis=-1),
            )
            vlines = _make_poisson_lines(mineral, n_rots, 0.02, plot_minimum=True)
            min_poisson = poisson[inds]
            prps.append(("Minimum Poisson ratio", (min_poisson, vlines)))
        elif type == "maximum poisson ratio":
            inds = np.ix_(*[np.arange(i) for i in poisson.shape[:-1]]) + (
                np.argmax(poisson, axis=-1),
            )
            max_poisson = poisson[inds]
            vlines = _make_poisson_lines(mineral, n_rots, 0.02, plot_minimum=False)
            prps.append(("Maximum Poisson ratio", (max_poisson, vlines)))
        else:
            raise Exception("plot_type not recognised.")

    contour_sets = []
    ticks = []
    lines = []
    both_vs = "vs1" in plot_types and "vs2" in plot_types
    both_poisson = (
        "minimum poisson ratio" in plot_types and "maximum poisson ratio" in plot_types
    )
    if both_vs:
        vsmin = np.min(np.array([vs1s, vs2s])) / 1.0e3
        vsmax = np.max(np.array([vs1s, vs2s])) / 1.0e3
    if both_poisson:
        poissonmin = np.min(min_poisson)
        poissonmax = np.max(max_poisson)
    for i, prp in enumerate(prps):
        title, values = prp
        plot_lines = False
        if len(values) == 2:
            if plot_vectors is True:
                plot_lines = True
            vlines = values[1]
            values = values[0]

        axes[i].set_title(title)

        vmin = np.min(values)
        vmax = np.max(values)
        vmin2 = np.min(values)
        vmax2 = np.max(values)

        if (title == "V$_{S1}$ (km/s)" or title == "V$_{S2}$ (km/s)") and both_vs:
            vmin = vsmin
            vmax = vsmax
        if (
            title == "Minimum Poisson ratio" or title == "Maximum Poisson ratio"
        ) and both_poisson:
            vmin = poissonmin
            vmax = poissonmax

        spacing = np.power(10.0, np.floor(np.log10(vmax2 - vmin2)))
        nt = int((vmax2 - vmin2 - vmax2 % spacing + vmin2 % spacing) / spacing)
        if nt == 1:
            spacing = spacing / 4.0
        elif nt < 4:
            spacing = spacing / 2.0
        elif nt > 8:
            spacing = spacing * 2.0

        tmin = vmin2 + (spacing - vmin2 % spacing)
        tmax = vmax2 - vmax2 % spacing
        nt = int((tmax - tmin) / spacing + 1)

        ticks.append(np.linspace(tmin, tmax, nt))
        contour_sets.append(
            axes[i].contourf(
                theta, r, values, n_divs, cmap=plt.cm.jet_r, vmin=vmin, vmax=vmax
            )
        )
        lines.append(
            axes[i].contour(
                theta, r, values, ticks[-1], colors=("black",), linewidths=(1,)
            )
        )
        if plot_lines:
            for line in vlines:
                axes[i].plot(line[0], line[1], color="white", linewidth=1)
            axes[i].set_ylim(0.0, 1.0)

        axes[i].set_yticks([])

    return contour_sets, ticks, lines



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.

from burnman.eos.mie_grueneisen_debye import MGDBase
from burnman.classes.calibrant import Calibrant


"""
Matsui_2012
^^^^^^^^^^^
"""


class NaCl_B1(Calibrant):
    """
    The NaCl (B1 structured) pressure standard reported by
    Matsui (2012; https://doi.org/10.2138/am.2012.4136).
    """

    def __init__(self):
        def _pressure_Matsui_NaCl(volume, temperature, params):

            # Isothermal pressure (GPa)
            a = (3 / 2) * (params["Kprime_0"] - 4)
            b = (
                9 * params["K_0"] * params["Kprime_prime_0"]
                + 9 * params["Kprime_0"] ** 2
                - 63 * params["Kprime_0"]
                + 143
            ) / 6.0
            f = 0.5 * ((volume / params["V_0"]) ** (-2 / 3) - 1)
            K_T = (
                params["K_0"]
                * ((1 + 2 * f) ** (5 / 2))
                * (1 + (7 + 2 * a) * f + (9 * a + 3 * b) * f**2 + 11 * b * f**3)
            )
            P0 = 3 * f * params["K_0"] * (1 + 2 * f) ** (5 / 2) * (1 + a * f + b * f**2)

            # Thermal pressure
            thermal_model = MGDBase()
            Pth0 = thermal_model._thermal_pressure(params["T_0"], volume, params)
            Pth = thermal_model._thermal_pressure(temperature, volume, params)

            # Total pressure
            P = (P0 * 1e9) + Pth - Pth0

            return P

        _params_Matsui_NaCl = {
            "V_0": 2.7013e-05,
            "K_0": 23.7,
            "Kprime_0": 5.14,
            "Kprime_prime_0": -0.392,
            "Debye_0": 279.0,
            "grueneisen_0": 1.56,
            "q_0": 0.96,
            "n": 2.0,
            "T_0": 300.0,
            "P_0": 0.0,
            "Z": 4.0,
        }

        Calibrant.__init__(self, _pressure_Matsui_NaCl, "pressure", _params_Matsui_NaCl)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
Calibrant database

  - :mod:`~burnman.calibrants.Anderson_1989`
  - :mod:`~burnman.calibrants.Armentrout_2015`
  - :mod:`~burnman.calibrants.Campbell_2009`
  - :mod:`~burnman.calibrants.Chidester_2021`
  - :mod:`~burnman.calibrants.Decker_1971`
  - :mod:`~burnman.calibrants.Dewaele_2008`
  - :mod:`~burnman.calibrants.Dewaele_2012`
  - :mod:`~burnman.calibrants.Dewaele_2013`
  - :mod:`~burnman.calibrants.Dewaele_2020`
  - :mod:`~burnman.calibrants.Dorogokupets_2007`
  - :mod:`~burnman.calibrants.Dorogokupets_2017`
  - :mod:`~burnman.calibrants.Dubrovinsky_1998`
  - :mod:`~burnman.calibrants.Fei_2007`
  - :mod:`~burnman.calibrants.Fei_2016`
  - :mod:`~burnman.calibrants.Holmes_1989`
  - :mod:`~burnman.calibrants.Huang_2016`
  - :mod:`~burnman.calibrants.LeGodec_2000`
  - :mod:`~burnman.calibrants.Litasov_2013`
  - :mod:`~burnman.calibrants.Matsui_2009`
  - :mod:`~burnman.calibrants.Matsui_2010`
  - :mod:`~burnman.calibrants.Matsui_2012`
  - :mod:`~burnman.calibrants.Miozzi_2020`
  - :mod:`~burnman.calibrants.Ono_2022`
  - :mod:`~burnman.calibrants.Pigott_2015`
  - :mod:`~burnman.calibrants.Shi_2022`
  - :mod:`~burnman.calibrants.Shim_2002`
  - :mod:`~burnman.calibrants.Sokolova_2013`
  - :mod:`~burnman.calibrants.Speziale_2001`
  - :mod:`~burnman.calibrants.Tange_2009`
  - :mod:`~burnman.calibrants.Tateno_2019`
  - :mod:`~burnman.calibrants.Tsuchiya_2003`
  - :mod:`~burnman.calibrants.Walker_2002`
  - :mod:`~burnman.calibrants.Zeng_2010`
  - :mod:`~burnman.calibrants.Zha_2004`
  - :mod:`~burnman.calibrants.Zha_2008`
  - :mod:`~burnman.calibrants.Zhang_1999`
  - :mod:`~burnman.calibrants.Zhao_1997`
  - :mod:`~burnman.calibrants.Zhao_2000`
"""

from . import Anderson_1989
from . import Armentrout_2015
from . import Campbell_2009
from . import Chidester_2021
from . import Decker_1971
from . import Dewaele_2008
from . import Dewaele_2012
from . import Dewaele_2013
from . import Dewaele_2020
from . import Dorogokupets_2007
from . import Dorogokupets_2017
from . import Dubrovinsky_1998
from . import Fei_2007
from . import Fei_2016
from . import Holmes_1989
from . import Huang_2016
from . import LeGodec_2000
from . import Litasov_2013
from . import Matsui_2009
from . import Matsui_2010
from . import Matsui_2012
from . import Miozzi_2020
from . import Ono_2022
from . import Pigott_2015
from . import Shi_2022
from . import Shim_2002
from . import Sokolova_2013
from . import Speziale_2001
from . import Tange_2009
from . import Tateno_2019
from . import Tsuchiya_2003
from . import Walker_2002
from . import Zeng_2010
from . import Zha_2004
from . import Zha_2008
from . import Zhang_1999
from . import Zhao_1997
from . import Zhao_2000
from . import tools



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.

from burnman.eos.birch_murnaghan import BirchMurnaghanBase as BM3
from burnman.eos.mie_grueneisen_debye import MGDBase
from burnman.classes.calibrant import Calibrant


"""
Huang_2016
^^^^^^^^^^
"""


class Mo(Calibrant):
    """
    The Mo pressure standard reported by
    Huang et al. (2016; https://doi.org/10.1038/srep19923).
    """

    def __init__(self):
        def _pressure_Huang_Mo(volume, temperature, params):

            # Isothermal pressure (GPa)
            pressure_model = BM3()
            P0 = pressure_model.pressure(params["T_0"], volume, params)

            # Thermal pressure
            thermal_model = MGDBase()
            Pth0 = thermal_model._thermal_pressure(params["T_0"], volume, params)
            Pth = thermal_model._thermal_pressure(temperature, volume, params)

            # Total pressure
            P = P0 + Pth - Pth0

            return P

        _params_Huang_Mo = {
            "V_0": 9.37647e-06,
            "K_0": 255.0e9,
            "Kprime_0": 4.25,
            "Debye_0": 470.0,
            "grueneisen_0": 2.01,
            "q_0": 0.6,
            "n": 1.0,
            "T_0": 300.0,
            "P_0": 0.0,
            "Z": 2.0,
        }

        Calibrant.__init__(self, _pressure_Huang_Mo, "pressure", _params_Huang_Mo)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

# Import common lower level function for backwards compatibility
from ..utils.unitcell import molar_volume_from_unit_cell_volume



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.


from __future__ import absolute_import

import numpy as np
from sympy import Matrix, nsimplify
from collections import OrderedDict

from .material import material_property, cached_property
from .mineral import Mineral
from .solutionmodel import MechanicalSolution
from .solutionmodel import PolynomialSolution
from .averaging_schemes import reuss_average_function

from ..utils.reductions import independent_row_indices
from ..utils.chemistry import sum_formulae, sort_element_list_to_IUPAC_order
from ..utils.misc import copy_documentation
from ..utils.math import complete_basis

import copy
from scipy.optimize import minimize


class Solution(Mineral):
    """
    This is the base class for all solutions.
    Site occupancies, endmember activities and the constant
    and pressure and temperature dependencies of the excess
    properties can be queried after using set_composition().
    States of the solution can only be queried after setting
    the pressure, temperature and composition using set_state().

    This class is available as :class:`burnman.Solution`.
    It uses an instance of :class:`burnman.SolutionModel` to
    calculate interaction terms between endmembers.

    All the solution parameters are expected to be in SI units.  This
    means that the interaction parameters should be in J/mol, with the T
    and P derivatives in J/K/mol and m^3/mol.

    The parameters are relevant to all solution models. Please
    see the documentation for individual models for details about
    other parameters.

    :param name: Name of the solution.
    :type name: string
    :param solution_model: The SolutionModel object defining the properties
        of the solution.
    :type solution_model: :class:`burnman.SolutionModel`
    :param molar_fractions: The molar fractions of each endmember
        in the solution. Can be reset using the set_composition() method.
    :type molar_fractions: numpy.array
    """

    def __init__(self, name=None, solution_model=None, molar_fractions=None):
        """
        Set up matrices to speed up calculations for when P, T, X is defined.
        """
        Mineral.__init__(self)

        # Solution needs a method attribute to call Mineral.set_state().
        # Note that set_method() below will not change self.method
        self.method = "SolutionMethod"

        if name is not None:
            self.name = name
        if solution_model is not None:
            self.solution_model = solution_model

        # Equation of state
        for mbr in self.solution_model.endmembers:
            mbr[0].set_method(mbr[0].params["equation_of_state"])

        # Molar fractions
        if molar_fractions is not None:
            self.set_composition(molar_fractions)

    @cached_property
    def endmembers(self):
        return self.solution_model.endmembers

    def set_composition(self, molar_fractions):
        """
        Set the composition for this solution.
        Resets cached properties.

        :param molar_fractions: Molar abundance for each endmember,
            needs to sum to one.
        :type molar_fractions: list of float
        """
        assert len(self.solution_model.endmembers) == len(molar_fractions)

        if type(self.solution_model) is not MechanicalSolution:
            if np.abs(sum(molar_fractions) - 1.0) > 1.0e-4:
                raise ValueError(
                    "Molar fractions do not sum to one for "
                    "the current instance of "
                    f"<{self.name}>: {molar_fractions}"
                )

        if type(self.solution_model) is PolynomialSolution:
            self.solution_model.set_composition(molar_fractions)

        self.reset()
        self.molar_fractions = np.array(molar_fractions)

    def set_method(self, method):
        for i in range(self.n_endmembers):
            self.solution_model.endmembers[i][0].set_method(method)
        # note: do not set self.method here!
        self.reset()

    def set_state(self, pressure, temperature):
        if type(self.solution_model) is PolynomialSolution:
            self.solution_model.set_state(pressure, temperature)

        Mineral.set_state(self, pressure, temperature)
        for mbr in self.solution_model.endmembers:
            mbr[0].set_state(pressure, temperature)

    @material_property
    def formula(self):
        """
        Returns molar chemical formula of the solution.
        :rtype: Counter
        """
        return sum_formulae(self.endmember_formulae, self.molar_fractions)

    @material_property
    def site_occupancies(self):
        """
        :returns: The fractional occupancies of species on each site.
        :rtype: list of OrderedDicts
        """
        f = self.molar_fractions
        occs = np.einsum("ij, i", self.solution_model.endmember_occupancies, f)
        site_occs = []
        k = 0
        for site in self.solution_model.sites:
            site_occs.append(OrderedDict())
            for species in site:
                site_occs[-1][species] = occs[k]
                k += 1

        return site_occs

    def site_formula(self, precision=2):
        """
        Returns the molar chemical formula of the solution with
            site occupancies. For example, [Mg0.4Fe0.6]2SiO4.

        :param precision: Precision with which to print the site occupancies
        :type precision: int

        :returns: Molar chemical formula of the solution with site occupancies
        :rtype: str
        """
        split_empty = self.solution_model.empty_formula.split("[")
        formula = split_empty[0]
        for i, site_occs in enumerate(self.site_occupancies):
            formula += "["
            for species, occ in site_occs.items():
                if np.abs(occ) < 1.0e-12:
                    occ = np.abs(occ)
                formula += f"{species}{occ:0.{precision}f}"
            formula += split_empty[i + 1]
        return formula

    @material_property
    def activities(self):
        """
        Returns a list of endmember activities [unitless].
        """
        return self.solution_model.activities(
            self.pressure, self.temperature, self.molar_fractions
        )

    @material_property
    def activity_coefficients(self):
        """
        Returns a list of endmember activity coefficients
        (gamma = activity / ideal activity) [unitless].
        """
        return self.solution_model.activity_coefficients(
            self.pressure, self.temperature, self.molar_fractions
        )

    @material_property
    def molar_internal_energy(self):
        """
        Returns molar internal energy of the mineral [J/mol].
        Aliased with self.energy
        """
        return self.molar_helmholtz + self.temperature * self.molar_entropy

    @material_property
    def excess_partial_gibbs(self):
        """
        Returns excess partial molar gibbs free energy [J/mol].
        Property specific to solutions.
        """
        return self.solution_model.excess_partial_gibbs_free_energies(
            self.pressure, self.temperature, self.molar_fractions
        )

    @material_property
    def excess_partial_volumes(self):
        """
        Returns excess partial volumes [m^3].
        Property specific to solutions.
        """
        return self.solution_model.excess_partial_volumes(
            self.pressure, self.temperature, self.molar_fractions
        )

    @material_property
    def excess_partial_entropies(self):
        """
        Returns excess partial entropies [J/K].
        Property specific to solutions.
        """
        return self.solution_model.excess_partial_entropies(
            self.pressure, self.temperature, self.molar_fractions
        )

    @material_property
    def partial_gibbs(self):
        """
        Returns endmember partial molar gibbs free energy [J/mol].
        Property specific to solutions.
        """
        return (
            np.array([mbr[0].gibbs for mbr in self.solution_model.endmembers])
            + self.excess_partial_gibbs
        )

    @material_property
    def partial_volumes(self):
        """
        Returns endmember partial volumes [m^3].
        Property specific to solutions.
        """
        return (
            np.array([mbr[0].molar_volume for mbr in self.solution_model.endmembers])
            + self.excess_partial_volumes
        )

    @material_property
    def partial_entropies(self):
        """
        Returns endmember partial entropies [J/K].
        Property specific to solutions.
        """
        return (
            np.array([mbr[0].molar_entropy for mbr in self.solution_model.endmembers])
            + self.excess_partial_entropies
        )

    @material_property
    def excess_gibbs(self):
        """
        Returns molar excess gibbs free energy [J/mol].
        Property specific to solutions.
        """
        return self.solution_model.excess_gibbs_free_energy(
            self.pressure, self.temperature, self.molar_fractions
        )

    @material_property
    def gibbs_hessian(self):
        """
        Returns an array containing the second compositional derivative
        of the Gibbs free energy [J]. Property specific to solutions.
        """
        return self.solution_model.gibbs_hessian(
            self.pressure, self.temperature, self.molar_fractions
        )

    @material_property
    def entropy_hessian(self):
        """
        Returns an array containing the second compositional derivative
        of the entropy [J/K]. Property specific to solutions.
        """
        return self.solution_model.entropy_hessian(
            self.pressure, self.temperature, self.molar_fractions
        )

    @material_property
    def volume_hessian(self):
        """
        Returns an array containing the second compositional derivative
        of the volume [m^3]. Property specific to solutions.
        """
        return self.solution_model.volume_hessian(
            self.pressure, self.temperature, self.molar_fractions
        )

    @material_property
    def molar_gibbs(self):
        """
        Returns molar Gibbs free energy of the solution [J/mol].
        Aliased with self.gibbs.
        """
        f = self.molar_fractions
        return (
            sum(
                [
                    self.solution_model.endmembers[i][0].gibbs * f[i]
                    for i in range(self.n_endmembers)
                ]
            )
            + self.excess_gibbs
        )

    @material_property
    def molar_helmholtz(self):
        """
        Returns molar Helmholtz free energy of the solution [J/mol].
        Aliased with self.helmholtz.
        """
        return self.molar_gibbs - self.pressure * self.molar_volume

    @material_property
    def molar_mass(self):
        """
        Returns molar mass of the solution [kg/mol].
        """
        return sum(
            [
                self.solution_model.endmembers[i][0].molar_mass
                * self.molar_fractions[i]
                for i in range(self.n_endmembers)
            ]
        )

    @material_property
    def excess_volume(self):
        """
        Returns excess molar volume of the solution [m^3/mol].
        Specific property for solutions.
        """
        return self.solution_model.excess_volume(
            self.pressure, self.temperature, self.molar_fractions
        )

    @material_property
    def molar_volume(self):
        """
        Returns molar volume of the solution [m^3/mol].
        Aliased with self.V.
        """
        return (
            sum(
                [
                    self.solution_model.endmembers[i][0].molar_volume
                    * self.molar_fractions[i]
                    for i in range(self.n_endmembers)
                ]
            )
            + self.excess_volume
        )

    @material_property
    def density(self):
        """
        Returns density of the solution [kg/m^3].
        Aliased with self.rho.
        """
        return self.molar_mass / self.molar_volume

    @material_property
    def excess_entropy(self):
        """
        Returns excess molar entropy [J/K/mol].
        Property specific to solutions.
        """
        return self.solution_model.excess_entropy(
            self.pressure, self.temperature, self.molar_fractions
        )

    @material_property
    def molar_entropy(self):
        """
        Returns molar entropy of the solution [J/K/mol].
        Aliased with self.S.
        """
        f = self.molar_fractions
        return (
            sum(
                [
                    self.solution_model.endmembers[i][0].S * f[i]
                    for i in range(self.n_endmembers)
                ]
            )
            + self.excess_entropy
        )

    @material_property
    def excess_enthalpy(self):
        """
        Returns excess molar enthalpy [J/mol].
        Property specific to solutions.
        """
        return self.solution_model.excess_enthalpy(
            self.pressure, self.temperature, self.molar_fractions
        )

    @material_property
    def molar_enthalpy(self):
        """
        Returns molar enthalpy of the solution [J/mol].
        Aliased with self.H.
        """
        f = self.molar_fractions
        return (
            sum(
                [
                    self.solution_model.endmembers[i][0].H * f[i]
                    for i in range(self.n_endmembers)
                ]
            )
            + self.excess_enthalpy
        )

    @material_property
    def isothermal_bulk_modulus_reuss(self):
        """
        Returns isothermal bulk modulus of the solution [Pa].
        Aliased with self.K_T.
        """
        return (
            self.V
            * 1.0
            / (
                sum(
                    [
                        self.solution_model.endmembers[i][0].V
                        / (
                            self.solution_model.endmembers[i][
                                0
                            ].isothermal_bulk_modulus_reuss
                        )
                        * self.molar_fractions[i]
                        for i in range(self.n_endmembers)
                    ]
                )
                + self.solution_model.VoverKT_excess()
            )
        )

    @material_property
    def isentropic_bulk_modulus_reuss(self):
        """
        Returns adiabatic bulk modulus of the solution [Pa].
        Aliased with self.K_S.
        """
        if self.temperature < 1e-10:
            return self.isothermal_bulk_modulus_reuss
        else:
            return (
                self.isothermal_bulk_modulus_reuss
                * self.molar_heat_capacity_p
                / self.molar_heat_capacity_v
            )

    @material_property
    def isothermal_compressibility_reuss(self):
        """
        Returns isothermal compressibility of the solution.
        (or inverse isothermal bulk modulus) [1/Pa].
        Aliased with self.K_T.
        """
        return 1.0 / self.isothermal_bulk_modulus_reuss

    @material_property
    def isentropic_compressibility_reuss(self):
        """
        Returns adiabatic compressibility of the solution.
        (or inverse adiabatic bulk modulus) [1/Pa].
        Aliased with self.K_S.
        """
        return 1.0 / self.isentropic_bulk_modulus_reuss

    @material_property
    def shear_modulus(self):
        """
        Returns shear modulus of the solution [Pa].
        Aliased with self.G.
        """
        G_list = np.fromiter(
            (e[0].G for e in self.solution_model.endmembers),
            dtype=float,
            count=self.n_endmembers,
        )
        return reuss_average_function(self.molar_fractions, G_list)

    @material_property
    def p_wave_velocity(self):
        """
        Returns P wave speed of the solution [m/s].
        Aliased with self.v_p.
        """
        return np.sqrt(
            (self.isentropic_bulk_modulus_reuss + 4.0 / 3.0 * self.shear_modulus)
            / self.density
        )

    @material_property
    def bulk_sound_velocity(self):
        """
        Returns bulk sound speed of the solution [m/s].
        Aliased with self.v_phi.
        """
        return np.sqrt(self.isentropic_bulk_modulus_reuss / self.density)

    @material_property
    def shear_wave_velocity(self):
        """
        Returns shear wave speed of the solution [m/s].
        Aliased with self.v_s.
        """
        return np.sqrt(self.shear_modulus / self.density)

    @material_property
    def grueneisen_parameter(self):
        """
        Returns grueneisen parameter of the solution [unitless].
        Aliased with self.gr.
        """
        if self.temperature < 1e-10:
            return float("nan")
        else:
            return (
                self.thermal_expansivity
                * self.isothermal_bulk_modulus_reuss
                * self.molar_volume
                / self.molar_heat_capacity_v
            )

    @material_property
    def thermal_expansivity(self):
        """
        Returns thermal expansion coefficient (alpha)
        of the solution [1/K].
        Aliased with self.alpha.
        """
        return (1.0 / self.V) * (
            sum(
                [
                    self.solution_model.endmembers[i][0].alpha
                    * self.solution_model.endmembers[i][0].V
                    * self.molar_fractions[i]
                    for i in range(self.n_endmembers)
                ]
            )
            + self.solution_model.alphaV_excess()
        )

    @material_property
    def molar_heat_capacity_v(self):
        """
        Returns molar heat capacity at constant volume of the
        solution [J/K/mol].
        Aliased with self.C_v.
        """
        return (
            self.molar_heat_capacity_p
            - self.molar_volume
            * self.temperature
            * self.thermal_expansivity
            * self.thermal_expansivity
            * self.isothermal_bulk_modulus_reuss
        )

    @material_property
    def molar_heat_capacity_p(self):
        """
        Returns molar heat capacity at constant pressure
        of the solution [J/K/mol].
        Aliased with self.C_p.
        """
        return (
            sum(
                [
                    self.solution_model.endmembers[i][0].molar_heat_capacity_p
                    * self.molar_fractions[i]
                    for i in range(self.n_endmembers)
                ]
            )
            + self.solution_model.Cp_excess()
        )

    @cached_property
    def stoichiometric_matrix(self):
        """
        A sympy Matrix where each element M[i,j] corresponds
        to the number of atoms of element[j] in endmember[i].
        """

        def f(i, j):
            e = self.elements[j]
            if e in self.endmember_formulae[i]:
                return nsimplify(self.endmember_formulae[i][e])
            else:
                return 0

        return Matrix(len(self.endmember_formulae), len(self.elements), f)

    @cached_property
    def stoichiometric_array(self):
        """
        An array where each element arr[i,j] corresponds
        to the number of atoms of element[j] in endmember[i].
        """
        return np.array(self.stoichiometric_matrix)

    @cached_property
    def reaction_basis(self):
        """
        An array where each element arr[i,j] corresponds
        to the number of moles of endmember[j] involved in reaction[i].
        """
        reaction_basis = np.array(
            [v[:] for v in self.stoichiometric_matrix.T.nullspace()], dtype=float
        )

        if len(reaction_basis) == 0:
            reaction_basis = np.empty((0, len(self.endmember_names)))

        return reaction_basis

    @cached_property
    def n_reactions(self):
        """
        The number of reactions in reaction_basis.
        """
        return len(self.reaction_basis[:, 0])

    @cached_property
    def compositional_basis(self):
        """_summary_

        :return: _description_
        :rtype: _type_
        """
        return complete_basis(self.reaction_basis)[self.n_reactions :]

    @cached_property
    def independent_element_indices(self):
        """
        A list of an independent set of element indices. If the amounts of
        these elements are known (element_amounts),
        the amounts of the other elements can be inferred by
        -compositional_null_basis[independent_element_indices].dot(element_amounts).
        """
        return sorted(independent_row_indices(self.stoichiometric_matrix.T))

    @cached_property
    def dependent_element_indices(self):
        """
        The element indices not included in the independent list.
        """
        return [
            i
            for i in range(len(self.elements))
            if i not in self.independent_element_indices
        ]

    @cached_property
    def compositional_null_basis(self):
        """
        An array N such that N.b = 0 for all bulk compositions that can
        be produced with a linear sum of the endmembers in the solution.
        """
        null = self.stoichiometric_matrix.nullspace()
        null_basis = np.array([v[:] for v in null])

        M = null_basis[:, self.dependent_element_indices]
        assert M.shape[0] == M.shape[1]
        assert (np.abs(M - np.eye(M.shape[0])) < 1.0e-20).all()

        return null_basis

    @cached_property
    def endmember_formulae(self):
        """
        A list of formulae for all the endmember in the solution.
        """
        mbrs = self.solution_model.endmembers
        return [mbr[0].params["formula"] for mbr in mbrs]

    @cached_property
    def endmember_names(self):
        """
        A list of names for all the endmember in the solution.
        """
        return [mbr[0].name for mbr in self.solution_model.endmembers]

    @cached_property
    def n_endmembers(self):
        """
        The number of endmembers in the solution.
        """
        return len(self.solution_model.endmembers)

    @cached_property
    def elements(self):
        """
        A list of the elements which could be contained in the solution,
        returned in the IUPAC element order.
        """
        keys = []
        for f in self.endmember_formulae:
            keys.extend(f.keys())

        return sort_element_list_to_IUPAC_order(set(keys))


SolidSolution = Solution


class RelaxedSolution(Solution):
    """
    A class implementing a relaxed solution
    model.
    This class is derived from Solution,
    and inherits most of the methods from that class.

    Instantiation of a RelaxedSolution involves passing
    an Solution, plus a set of vectors that represent rapid
    deformation modes. For example, a solution of MgO, FeHSO and FeLSO
    (high and low spin wuestite) can rapidly change proportion of
    high spin and low spin iron, and so a single vector should be passed:
    np.array([[0., -1., 1.]]) or some multiple thereof.

    States of the mineral can only be queried after setting the
    pressure and temperature using set_state() and the composition using
    set_composition().

    This class is available as ``burnman.RelaxedSolution``.
    """

    def __init__(
        self,
        solution,
        relaxation_vectors,
        unrelaxed_vectors,
    ):
        # Make an attribute with the unrelaxed solution
        self.unrelaxed = solution

        # The relaxation vectors
        self.n_relaxation_vectors = len(relaxation_vectors)
        self.n_unrelaxed_vectors = len(unrelaxed_vectors)

        self.dndq = np.array(relaxation_vectors).T
        assert len(self.dndq.shape) == 2
        assert len(self.dndq) == self.unrelaxed.n_endmembers

        self.dndx = np.array(unrelaxed_vectors).T
        assert len(self.dndx.shape) == 2
        assert len(self.dndx) == self.unrelaxed.n_endmembers
        assert (
            len(unrelaxed_vectors) + len(relaxation_vectors)
            == self.unrelaxed.n_endmembers
        )

        self.q_initial = np.zeros(len(relaxation_vectors)) + 0.001

        try:
            molar_fractions = solution.molar_fractions
        except AttributeError:
            molar_fractions = None

        # Give the relaxed solution the same base properties as the
        # unrelaxed solution
        Solution.__init__(
            self,
            name=solution.name,
            solution_model=solution.solution_model,
            molar_fractions=molar_fractions,
        )

    def set_state(self, pressure, temperature, relaxed=True):
        """
        Sets the state of the solution. Also relaxes the
        structure parameters if set_composition() has already
        been used and if the relaxed argument has been set to
        True.

        :param pressure: The pressure of the solution [Pa]
        :type pressure: float
        :param temperature: The temperature of the solution [K]
        :type temperature: float
        :param relaxed: Whether to minimize the Gibbs energy
            of the material by changing the values of the structure
            parameters. Defaults to True.
        :type relaxed: bool, optional
        """
        self.unrelaxed.set_state(pressure, temperature)

        if hasattr(self.unrelaxed, "molar_fractions") and relaxed:
            self._relax_at_PTX()
            Solution.set_state(self, pressure, temperature)

    def set_composition(self, molar_fractions, q_initial=None, relaxed=True):
        """
        Sets the composition of the model. Also relaxes the
        structure parameters if set_state() has already
        been used and if the relaxed argument has been set to
        True.

        :param molar_fractions: Molar fractions of the
            independent endmembers corresponding to the
            unrelaxed vectors specified during initialisation.
        :type molar_fractions: 1D numpy array
        :param q_initial: Initial values of the structure parameters.
            Defaults to None, in which case the preexisting
            initial values are used
            (first set to 0.001 during initialisation).
        :type q_initial: 1D numpy array, optional
        :param relaxed: Whether to minimize the Gibbs energy
            of the material by changing the values of the structure
            parameters. Defaults to True.
        :type relaxed: bool, optional
        """
        self.unrelaxed_vectors = molar_fractions
        if q_initial is not None:
            self.q_initial = np.array(q_initial)
        n = np.einsum("ij, j", self.dndq, self.q_initial) + np.einsum(
            "ij, j", self.dndx, molar_fractions
        )

        self.unrelaxed.set_composition(n)

        if self.unrelaxed.pressure is not None and relaxed:
            self._relax_at_PTX()
            Solution.set_composition(self, n)
            self.unrelaxed.set_composition(self.molar_fractions)

    def _relax_at_PTX(self):
        """
        Minimizes the Gibbs energy at constant pressure and
        temperature by changing the structural parameters.

        Run during set_state() and set_composition(), as long as both
        state and composition have already been set. This function
        should not generally be needed by the user.
        """
        n0 = copy.copy(self.unrelaxed.molar_fractions)

        def G_func(dq):
            n = n0 + np.einsum("ij, j", self.dndq, dq)
            self.unrelaxed.set_composition(n)
            return self.unrelaxed.molar_gibbs

        sol = minimize(G_func, np.zeros(len(self.dndq[0])), method="Nelder-Mead")
        assert sol.success
        n = n0 + np.einsum("ij, j", self.dndq, sol.x)
        self.unrelaxed.set_composition(n)
        Solution.set_composition(self, n)

    def set_state_with_volume(
        self, volume, temperature, pressure_guesses=[0.0e9, 10.0e9]
    ):
        """
        This function acts similarly to set_state, but takes volume and
        temperature as input to find the pressure. In order to ensure
        self-consistency, this function does not use any pressure functions
        from the material classes, but instead finds the pressure using the
        brentq root-finding and Nelder-Mead minimization methods.

        Composition should have been set before this function is used.

        :param volume: The desired molar volume of the mineral [m^3].
        :type volume: float

        :param temperature: The desired temperature of the mineral [K].
        :type temperature: float

        :param pressure_guesses: A list of floats denoting the initial
            low and high guesses for bracketing of the pressure [Pa].
            These guesses should preferably bound the correct pressure,
            but do not need to do so. More importantly,
            they should not lie outside the valid region of
            the equation of state. Defaults to [0.e9, 10.e9].
        :type pressure_guesses: list
        """
        ss = self.unrelaxed
        n0 = copy.copy(ss.molar_fractions)

        # Initial set_state without changing structural parameters
        # to estimate pressure
        ss.set_state_with_volume(volume, temperature, pressure_guesses)
        # Store in a mutable so that it can be updated in the loop
        pressure = [ss.pressure]

        def F_func(dq):
            n = n0 + np.einsum("ij, j", self.dndq, dq)
            ss.set_composition(n)
            ss.set_state_with_volume(
                volume, temperature, [pressure[0] - 1.0e9, pressure[0] + 1.0e9]
            )
            pressure[0] = ss.pressure
            return ss.molar_helmholtz

        sol = minimize(F_func, np.zeros(len(self.dndq[0])), method="Nelder-Mead")
        assert sol.success

        # Run one more time with root
        F_func(sol.x)
        self.unrelaxed.set_composition(ss.molar_fractions)
        Solution.set_composition(self, ss.molar_fractions)
        self.set_state(pressure[0], temperature)

    @material_property
    def _d2Gdqdq_fixed_PT(self):
        """
        Gibbs structural hessian
        calculated at constant pressure and temperature.
        """
        return np.einsum(
            "ij, ik, jl->kl", self.unrelaxed.gibbs_hessian, self.dndq, self.dndq
        )

    @material_property
    def _d2Gdqdz(self):
        """
        Second derivatives of the Helmholtz energy with
        respect to composition and z (pressure or temperature).
        """
        dVdq = np.einsum("i, ij->j", self.unrelaxed.partial_volumes, self.dndq)
        dSdq = np.einsum("i, ij->j", self.unrelaxed.partial_entropies, self.dndq)

        return np.array([dVdq, -dSdq]).T

    @material_property
    def _d2Gdqdq_fixed_PT_pinv(self):
        """
        The second derivative of the Helmholtz energy
        with respect to the structure parameters at constant
        strain and temperature. Often referred to as the
        susceptibility matrix.
        """
        return np.linalg.pinv(self._d2Gdqdq_fixed_PT)

    @material_property
    def dqdz_relaxed(self):
        """
        The change of the structure parameters with respect to
        strain and temperature that minimizes the Helmholtz
        energy.
        """
        return -np.einsum("kl, lj->kj", self._d2Gdqdq_fixed_PT_pinv, self._d2Gdqdz)

    @material_property
    def _d2Gdzdz_Q(self):
        """
        Block matrix of -V*beta_TR, V*alpha, -c_p/T
        at fixed Q
        """
        beta_TR = self.unrelaxed.isothermal_compressibility_reuss
        alpha = self.unrelaxed.thermal_expansivity
        c_p = self.unrelaxed.molar_heat_capacity_p
        V = self.molar_volume
        T = self.temperature
        return np.array([[-V * beta_TR, V * alpha], [V * alpha, -c_p / T]])

    @material_property
    def _d2Gdzdz(self):
        """
        Block matrix of -V*beta_TR, V*alpha, -c_p/T
        under Gibbs-minimizing Q
        """
        return self._d2Gdzdz_Q + np.einsum(
            "ki, kj->ij", self._d2Gdqdz, self.dqdz_relaxed
        )

    # The following scalar properties are all
    # second derivatives that are calculated in AnisotropicMineral
    # or Solution (rather than being derived from other properties),
    # and must therefore be redefined in relaxed materials

    # The volumetric molar heat capacity and grueneisen parameter are
    # derived properties in AnisotropicMineral
    # and so do not need to be redefined.

    @material_property
    @copy_documentation(Solution.isothermal_compressibility_reuss)
    def isothermal_compressibility_reuss(self):
        return -self._d2Gdzdz[0, 0] / self.V

    @material_property
    @copy_documentation(Solution.thermal_expansivity)
    def thermal_expansivity(self):
        return self._d2Gdzdz[0, 1] / self.V

    @material_property
    @copy_documentation(Solution.molar_heat_capacity_p)
    def molar_heat_capacity_p(self):
        return -self._d2Gdzdz[1, 1] * self.T

    @material_property
    @copy_documentation(Solution.isothermal_bulk_modulus_reuss)
    def isothermal_bulk_modulus_reuss(self):
        return 1.0 / self.isothermal_compressibility_reuss

    @material_property
    @copy_documentation(Solution.molar_heat_capacity_v)
    def molar_heat_capacity_v(self):
        return (
            self.molar_heat_capacity_p
            - self.molar_volume
            * self.temperature
            * self.thermal_expansivity
            * self.thermal_expansivity
            * self.isothermal_bulk_modulus_reuss
        )

    @material_property
    @copy_documentation(Solution.isentropic_bulk_modulus_reuss)
    def isentropic_bulk_modulus_reuss(self):
        if self.temperature < 1.0e-10:
            return self.isothermal_bulk_modulus_reuss
        else:
            return (
                self.isothermal_bulk_modulus_reuss
                * self.molar_heat_capacity_p
                / self.molar_heat_capacity_v
            )

    @material_property
    @copy_documentation(Solution.isentropic_compressibility_reuss)
    def isentropic_compressibility_reuss(self):
        return 1.0 / self.isentropic_bulk_modulus_reuss



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import
import numpy as np

"""
Functions for the Bukowinski (1977) model for the electronic component
of the Helmholtz energy as used by Stixrude an Lithgow-Bertelloni (2024).
"""


def helmholtz(temperature, volume, T_0, V_0, bel_0, gel):
    return (
        -0.5
        * bel_0
        * np.power(volume / V_0, gel)
        * (temperature * temperature - T_0 * T_0)
    )


def pressure(temperature, volume, T_0, V_0, bel_0, gel):
    """
    P = -dF/dV
    """
    return (
        0.5
        * gel
        * bel_0
        * np.power(volume / V_0, gel)
        * (temperature * temperature - T_0 * T_0)
        / volume
    )


def KToverV(temperature, volume, T_0, V_0, bel_0, gel):
    """
    KT = -V dP/dV
    """
    return -(gel - 1.0) * pressure(temperature, volume, T_0, V_0, bel_0, gel) / volume


def entropy(temperature, volume, V_0, bel_0, gel):
    """
    S = -dF/dT
    """
    return bel_0 * np.power(volume / V_0, gel) * temperature


def CVoverT(volume, V_0, bel_0, gel):
    """
    CV = T dS/dT
    """
    return bel_0 * np.power(volume / V_0, gel)


def aKT(temperature, volume, V_0, bel_0, gel):
    """
    aKT = dP/dT
    """
    return gel * bel_0 * np.power(volume / V_0, gel) * temperature / volume



from __future__ import absolute_import

# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences.
# Copyright (C) 2012 - 2024 by the BurnMan team, released under the GNU
# GPL v2 or later.


import scipy.optimize as opt
from . import equation_of_state as eos
import warnings
import numpy as np


# Try to import the jit from numba.  If it is
# not available, just go with the standard
# python interpreter
try:
    from numba import jit
except ImportError:

    def jit(fn):
        return fn


@jit(nopython=True)
def make_params(K0, K0_prime, K_infinity_prime):
    a = (
        16.0 * np.power(K0_prime, 3.0)
        + 84.0 * np.power(K0_prime, 2.0)
        + 192.0 * K0_prime
        - 972.0 * K_infinity_prime
        + 1177.0
    )
    b = 2.0 * np.power(K0_prime, 2.0) + 7.0 * K0_prime - 27.0 * K_infinity_prime + 38.0
    omega = np.power((a + np.sqrt(a * a - 32.0 * b * b * b)), 1.0 / 3.0)
    C = (
        (11.0 / 6.0)
        + (1.0 / 3.0) * K0_prime
        - K_infinity_prime
        + (np.power(2, -1.0 / 3.0) / 6) * omega
        + (np.power(2, 1.0 / 3.0) / 3) * (b / omega)
    )
    B = K_infinity_prime - 1.0
    A = K0 / (B - 0.5 * C + np.power(B + C, 2.0))
    return A, B, C


class MACAW(eos.EquationOfState):
    """
    Class for the MACAW equation of state
    detailed in Lozano and Aslam (2022; https://doi.org/10.1063/5.0076897).

    This equation of state has no temperature dependence.
    """

    def isothermal_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus :math:`K_T` :math:`[Pa]` as a function of pressure :math:`[Pa]`,
        temperature :math:`[K]` and volume :math:`[m^3]`.
        """
        A, B, C = make_params(params["K_0"], params["Kprime_0"], params["Kprime_inf"])
        Vrel = volume / params["V_0"]
        term1 = A * np.power(Vrel, -(B + 1))
        term2 = np.exp((2.0 / 3.0) * C * (1 - np.power(Vrel, 1.5)))
        term3 = np.power(C * np.power(Vrel, 1.5) + B, 2.0) - (
            0.5 * C * np.power(Vrel, 1.5) - B
        )
        return term1 * term2 * term3

    def volume(self, pressure, temperature, params):
        """
        Get the Vinet volume at a reference temperature for a given
        pressure :math:`[Pa]`. Returns molar volume in :math:`[m^3]`
        """

        def delta_pressure(x):
            return self.pressure(0.0, x, params) - pressure

        V = opt.brentq(delta_pressure, 0.1 * params["V_0"], 1.5 * params["V_0"])
        return V

    def pressure(self, temperature, volume, params):
        """
        Returns pressure :math:`[Pa]` as a function of volume :math:`[m^3]`.
        """
        A, B, C = make_params(params["K_0"], params["Kprime_0"], params["Kprime_inf"])
        Vrel = volume / params["V_0"]
        term1 = A * np.power(Vrel, -(B + 1.0))
        term2 = np.exp((2.0 / 3.0) * C * (1.0 - np.power(Vrel, 1.5)))
        term3 = C * np.power(Vrel, 1.5) + B
        return term1 * term2 * term3 - A * (B + C) + params["P_0"]

    def molar_internal_energy(self, pressure, temperature, volume, params):
        """
        Returns the internal energy :math:`\\mathcal{E}` of the mineral. :math:`[J/mol]`
        """
        A, B, C = make_params(params["K_0"], params["Kprime_0"], params["Kprime_inf"])
        Vrel = volume / params["V_0"]
        I1 = -params["V_0"] * (
            np.power(Vrel, -B) * np.exp((2.0 / 3.0) * C * (1.0 - np.power(Vrel, 1.5)))
            - 1.0
        )
        I0 = (-A * (B + C) + params["P_0"]) * params["V_0"] * (Vrel - 1.0)
        return -A * I1 - I0

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the Gibbs free energy :math:`\\mathcal{G}` of the mineral. :math:`[J/mol]`
        """
        return (
            self.molar_internal_energy(pressure, temperature, volume, params)
            + pressure * volume
        )

    def isentropic_bulk_modulus_reuss(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus :math:`K_s` of the mineral. :math:`[Pa]`.
        """
        return self.isothermal_bulk_modulus_reuss(pressure, temperature, volume, params)

    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Returns shear modulus :math:`G` of the mineral. :math:`[Pa]`
        """
        return 1.0e99

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the molar entropy :math:`\\mathcal{S}` of the mineral. :math:`[J/K/mol]`
        """
        return 0.0

    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, return a very small number. :math:`[J/K/mol]`
        """
        return 1.0e-99

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, return a very small number. :math:`[J/K/mol]`
        """
        return 1.0e-99

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, return zero. :math:`[1/K]`
        """
        return 0.0

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, return zero. :math:`[unitless]`
        """
        return 0.0

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters.
        The value for :math:`K'_{\\infty}` is thermodynamically bounded
        between 5/3 and :math:`K'_0` :cite:`StaceyDavis2004`.
        """

        if "E_0" not in params:
            params["E_0"] = 0.0
        if "P_0" not in params:
            params["P_0"] = 1.0e5

        # Check that all the required keys are in the dictionary
        expected_keys = ["V_0", "K_0", "Kprime_0", "Kprime_inf"]
        for k in expected_keys:
            if k not in params:
                raise KeyError("params object missing parameter : " + k)

        # Finally, check that the values are reasonable.
        if params["P_0"] < 0.0:
            warnings.warn("Unusual value for P_0", stacklevel=2)
        if params["V_0"] < 1.0e-7 or params["V_0"] > 1.0e-3:
            warnings.warn("Unusual value for V_0", stacklevel=2)
        if params["K_0"] < 1.0e9 or params["K_0"] > 1.0e13:
            warnings.warn("Unusual value for K_0", stacklevel=2)
        if params["Kprime_0"] < 0.0 or params["Kprime_0"] > 10.0:
            warnings.warn("Unusual value for Kprime_0", stacklevel=2)
        if params["Kprime_inf"] < 1 + 45.0 / 29.0:
            warnings.warn("Value for Kprime_inf below recommended value", stacklevel=2)
        if params["Kprime_inf"] > params["Kprime_0"]:
            warnings.warn("Kprime_inf should be less than Kprime_0", stacklevel=2)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
HHPH_2013
^^^^^^^^^

Minerals from Holland et al. (2013) and references therein.
The values in this document are all in S.I. units,
unlike those in the original paper.
File autogenerated using HHPHdata_to_burnman.py.
"""
from __future__ import absolute_import

from ..classes.mineral import Mineral
from ..utils.chemistry import dictionarize_formula, formula_mass


"""
ENDMEMBERS
"""


class fo(Mineral):
    def __init__(self):
        formula = "Mg2SiO4"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "fo",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -2172450.0,
            "S_0": 95.1,
            "V_0": 4.366e-05,
            "Cp": [233.3, 0.001494, -603800.0, -1869.7],
            "a_0": 2.85e-05,
            "K_0": 1.285e11,
            "Kprime_0": 3.84,
            "Kdprime_0": -3e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 530.0}
        Mineral.__init__(self)


class fa(Mineral):
    def __init__(self):
        formula = "Fe2SiO4"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "fa",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -1477740.0,
            "S_0": 151.0,
            "V_0": 4.631e-05,
            "Cp": [201.1, 0.01733, -1960600.0, -900.9],
            "a_0": 2.82e-05,
            "K_0": 1.256e11,
            "Kprime_0": 4.68,
            "Kdprime_0": -3.7e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 640.0}
        Mineral.__init__(self)


class mwd(Mineral):
    def __init__(self):
        formula = "Mg2SiO4"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "mwd",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -2138080.0,
            "S_0": 93.9,
            "V_0": 4.051e-05,
            "Cp": [208.7, 0.003942, -1709500.0, -1302.8],
            "a_0": 2.37e-05,
            "K_0": 1.726e11,
            "Kprime_0": 3.84,
            "Kdprime_0": -2.2e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 620.0}
        Mineral.__init__(self)


class fwd(Mineral):
    def __init__(self):
        formula = "Fe2SiO4"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "fwd",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -1467920.0,
            "S_0": 146.0,
            "V_0": 4.321e-05,
            "Cp": [201.1, 0.01733, -1960600.0, -900.9],
            "a_0": 2.73e-05,
            "K_0": 1.69e11,
            "Kprime_0": 4.35,
            "Kdprime_0": -2.6e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 900.0}
        Mineral.__init__(self)


class mrw(Mineral):
    def __init__(self):
        formula = "Mg2SiO4"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "mrw",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -2126840.0,
            "S_0": 90.0,
            "V_0": 3.949e-05,
            "Cp": [213.3, 0.00269, -1410400.0, -1495.9],
            "a_0": 2.01e-05,
            "K_0": 1.781e11,
            "Kprime_0": 4.35,
            "Kdprime_0": -2.4e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 630.0}
        Mineral.__init__(self)


class frw(Mineral):
    def __init__(self):
        formula = "Fe2SiO4"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "frw",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -1471760.0,
            "S_0": 140.0,
            "V_0": 4.203e-05,
            "Cp": [166.8, 0.04261, -1705400.0, -541.4],
            "a_0": 2.22e-05,
            "K_0": 1.977e11,
            "Kprime_0": 4.92,
            "Kdprime_0": -2.5e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 710.0}
        Mineral.__init__(self)


class mpv(Mineral):
    def __init__(self):
        formula = "MgSiO3"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "mpv",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -1442310.0,
            "S_0": 62.6,
            "V_0": 2.445e-05,
            "Cp": [149.3, 0.002918, -2983000.0, -799.1],
            "a_0": 1.87e-05,
            "K_0": 2.51e11,
            "Kprime_0": 4.14,
            "Kdprime_0": -1.6e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 470.0}
        Mineral.__init__(self)


class fpv(Mineral):
    def __init__(self):
        formula = "FeSiO3"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "fpv",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -1082910.0,
            "S_0": 95.0,
            "V_0": 2.534e-05,
            "Cp": [133.2, 0.01083, -3661400.0, -314.7],
            "a_0": 1.87e-05,
            "K_0": 2.81e11,
            "Kprime_0": 4.14,
            "Kdprime_0": -1.6e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 760.0}
        Mineral.__init__(self)


class apv(Mineral):
    def __init__(self):
        formula = "AlAlO3"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "apv",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -1619990.0,
            "S_0": 51.8,
            "V_0": 2.54e-05,
            "Cp": [139.5, 0.00589, -2460600.0, -589.2],
            "a_0": 1.8e-05,
            "K_0": 2.03e11,
            "Kprime_0": 4.0,
            "Kdprime_0": -2e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 770.0}
        Mineral.__init__(self)


class npv(Mineral):
    def __init__(self):
        formula = "Na0.5Al0.5SiO3"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "npv",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -1365000.0,
            "S_0": 63.0,
            "V_0": 2.334e-05,
            "Cp": [135.0, 0.00846, -1850300.0, -600.8],
            "a_0": 1.8e-05,
            "K_0": 2.03e11,
            "Kprime_0": 4.0,
            "Kdprime_0": -2e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 10240.0}
        Mineral.__init__(self)


class cpv(Mineral):
    def __init__(self):
        formula = "CaSiO3"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "cpv",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -1533590.0,
            "S_0": 74.5,
            "V_0": 2.745e-05,
            "Cp": [159.3, 0.0, -967300.0, -1075.4],
            "a_0": 2e-05,
            "K_0": 2.36e11,
            "Kprime_0": 3.9,
            "Kdprime_0": -1.6e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 1090.0}
        Mineral.__init__(self)


class mak(Mineral):
    def __init__(self):
        formula = "MgSiO3"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "mak",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -1489610.0,
            "S_0": 59.3,
            "V_0": 2.635e-05,
            "Cp": [147.8, 0.002015, -2395000.0, -801.8],
            "a_0": 2.12e-05,
            "K_0": 2.11e11,
            "Kprime_0": 4.55,
            "Kdprime_0": -2.2e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 420.0}
        Mineral.__init__(self)


class fak(Mineral):
    def __init__(self):
        formula = "FeSiO3"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "fak",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -1142130.0,
            "S_0": 91.5,
            "V_0": 2.76e-05,
            "Cp": [100.3, 0.013328, -4364900.0, 419.8],
            "a_0": 2.12e-05,
            "K_0": 2.18e11,
            "Kprime_0": 4.55,
            "Kdprime_0": -2.2e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 9630.0}
        Mineral.__init__(self)


class maj(Mineral):
    def __init__(self):
        formula = "Mg4Si4O12"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "maj",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -6041550.0,
            "S_0": 260.2,
            "V_0": 0.00011457,
            "Cp": [713.6, -0.000997, -1158200.0, -6622.3],
            "a_0": 1.83e-05,
            "K_0": 1.6e11,
            "Kprime_0": 4.56,
            "Kdprime_0": -2.8e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 2260.0}
        Mineral.__init__(self)


class nagt(Mineral):
    def __init__(self):
        formula = "Mg2NaAlSiSi3O12"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "nagt",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -5985000.0,
            "S_0": 260.6,
            "V_0": 0.0001109,
            "Cp": [620.8, 0.0112, -3755900.0, -4421.3],
            "a_0": 2.1e-05,
            "K_0": 1.7e11,
            "Kprime_0": 4.0,
            "Kdprime_0": -2.3e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 5120.0}
        Mineral.__init__(self)


class py(Mineral):
    def __init__(self):
        formula = "Mg3Al2Si3O12"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "py",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -6281770.0,
            "S_0": 269.5,
            "V_0": 0.00011313,
            "Cp": [633.5, 0.0, -5196100.0, -4315.2],
            "a_0": 2.37e-05,
            "K_0": 1.743e11,
            "Kprime_0": 4.05,
            "Kdprime_0": -2.3e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 990.0}
        Mineral.__init__(self)


class alm(Mineral):
    def __init__(self):
        formula = "Fe3Al2Si3O12"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "alm",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -5260750.0,
            "S_0": 342.0,
            "V_0": 0.00011525,
            "Cp": [677.3, 0.0, -3772700.0, -5044.0],
            "a_0": 2.12e-05,
            "K_0": 1.9e11,
            "Kprime_0": 2.98,
            "Kdprime_0": -1.6e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 1200.0}
        Mineral.__init__(self)


class gr(Mineral):
    def __init__(self):
        formula = "Ca3Al2Si3O12"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "gr",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -6643050.0,
            "S_0": 255.0,
            "V_0": 0.00012535,
            "Cp": [626.0, 0.0, -5779200.0, -4002.9],
            "a_0": 2.2e-05,
            "K_0": 1.72e11,
            "Kprime_0": 5.53,
            "Kdprime_0": -3.2e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 1370.0}
        Mineral.__init__(self)


class en(Mineral):
    def __init__(self):
        formula = "Mg2Si2O6"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "en",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -3090100.0,
            "S_0": 132.5,
            "V_0": 6.262e-05,
            "Cp": [356.2, -0.00299, -596900.0, -3185.3],
            "a_0": 2.27e-05,
            "K_0": 1.059e11,
            "Kprime_0": 8.65,
            "Kdprime_0": -8.2e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 620.0}
        Mineral.__init__(self)


class cen(Mineral):
    def __init__(self):
        formula = "Mg2Si2O6"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "cen",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -3090990.0,
            "S_0": 132.0,
            "V_0": 6.264e-05,
            "Cp": [306.0, -0.003793, -3041700.0, -1852.1],
            "a_0": 2.11e-05,
            "K_0": 1.059e11,
            "Kprime_0": 8.65,
            "Kdprime_0": -8.2e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 620.0}
        Mineral.__init__(self)


class hen(Mineral):
    def __init__(self):
        formula = "Mg2Si2O6"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "hen",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -3082610.0,
            "S_0": 131.7,
            "V_0": 6.099e-05,
            "Cp": [356.2, -0.00299, -596900.0, -3185.3],
            "a_0": 2.26e-05,
            "K_0": 1.5e11,
            "Kprime_0": 5.5,
            "Kdprime_0": -3.6e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 620.0}
        Mineral.__init__(self)


class hfs(Mineral):
    def __init__(self):
        formula = "Fe2Si2O6"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "hfs",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -2380810.0,
            "S_0": 189.0,
            "V_0": 6.405e-05,
            "Cp": [398.7, -0.006579, 1290100.0, -4058.0],
            "a_0": 2.37e-05,
            "K_0": 1.5e11,
            "Kprime_0": 5.5,
            "Kdprime_0": -3.6e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 790.0}
        Mineral.__init__(self)


class fs(Mineral):
    def __init__(self):
        formula = "Fe2Si2O6"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "fs",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -2388760.0,
            "S_0": 189.9,
            "V_0": 6.592e-05,
            "Cp": [398.7, -0.006579, 1290100.0, -4058.0],
            "a_0": 3.26e-05,
            "K_0": 1.01e11,
            "Kprime_0": 4.08,
            "Kdprime_0": -4e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 750.0}
        Mineral.__init__(self)


class mgts(Mineral):
    def __init__(self):
        formula = "MgAl2SiO6"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "mgts",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -3196600.0,
            "S_0": 131.0,
            "V_0": 6.05e-05,
            "Cp": [371.4, -0.004082, -398400.0, -3547.1],
            "a_0": 2.17e-05,
            "K_0": 1.028e11,
            "Kprime_0": 8.55,
            "Kdprime_0": -8.3e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 690.0}
        Mineral.__init__(self)


class di(Mineral):
    def __init__(self):
        formula = "CaMgSi2O6"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "di",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -3201820.0,
            "S_0": 142.9,
            "V_0": 6.619e-05,
            "Cp": [314.5, 4.1e-05, -2745900.0, -2020.1],
            "a_0": 2.73e-05,
            "K_0": 1.192e11,
            "Kprime_0": 5.19,
            "Kdprime_0": -4.4e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 580.0}
        Mineral.__init__(self)


class hed(Mineral):
    def __init__(self):
        formula = "CaFeSi2O6"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "hed",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -2842120.0,
            "S_0": 175.0,
            "V_0": 6.795e-05,
            "Cp": [340.2, 0.000812, -1047800.0, -2646.7],
            "a_0": 2.38e-05,
            "K_0": 1.192e11,
            "Kprime_0": 3.97,
            "Kdprime_0": -3.3e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 880.0}
        Mineral.__init__(self)


class jd(Mineral):
    def __init__(self):
        formula = "NaAlSiO6"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "jd",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -3025290.0,
            "S_0": 133.5,
            "V_0": 6.04e-05,
            "Cp": [319.4, 0.003616, -1173900.0, -2469.5],
            "a_0": 2.1e-05,
            "K_0": 1.281e11,
            "Kprime_0": 3.81,
            "Kdprime_0": -3e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 1520.0}
        Mineral.__init__(self)


class cats(Mineral):
    def __init__(self):
        formula = "CaAl2SiO6"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "cats",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -3310110.0,
            "S_0": 135.0,
            "V_0": 6.356e-05,
            "Cp": [347.6, -0.006974, -1781600.0, -2757.5],
            "a_0": 2.08e-05,
            "K_0": 1.192e11,
            "Kprime_0": 5.19,
            "Kdprime_0": -4.4e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 750.0}
        Mineral.__init__(self)


class stv(Mineral):
    def __init__(self):
        formula = "SiO2"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "stv",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -876820.0,
            "S_0": 24.0,
            "V_0": 1.401e-05,
            "Cp": [68.1, 0.00601, -1978200.0, -82.1],
            "a_0": 1.58e-05,
            "K_0": 3.09e11,
            "Kprime_0": 4.6,
            "Kdprime_0": -1.5e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 420.0}
        Mineral.__init__(self)


class macf(Mineral):
    def __init__(self):
        formula = "MgAl2O4"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "macf",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -2246420.0,
            "S_0": 80.0,
            "V_0": 3.614e-05,
            "Cp": [200.0, 0.006252, -2996400.0, -888.4],
            "a_0": 1.93e-05,
            "K_0": 2.12e11,
            "Kprime_0": 4.0,
            "Kdprime_0": -1.7e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 1080.0}
        Mineral.__init__(self)


class mscf(Mineral):
    def __init__(self):
        formula = "Mg2SiO4"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "mscf",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -2061130.0,
            "S_0": 87.5,
            "V_0": 3.649e-05,
            "Cp": [213.3, 0.00269, -1410400.0, -1495.9],
            "a_0": 2.01e-05,
            "K_0": 1.85e11,
            "Kprime_0": 4.0,
            "Kdprime_0": -1.7e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 1340.0}
        Mineral.__init__(self)


class fscf(Mineral):
    def __init__(self):
        formula = "Fe2SiO4"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "fscf",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -1405500.0,
            "S_0": 143.4,
            "V_0": 3.914e-05,
            "Cp": [181.1, 0.018526, -2767200.0, -527.1],
            "a_0": 2.01e-05,
            "K_0": 1.85e11,
            "Kprime_0": 4.0,
            "Kdprime_0": -1.7e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 10240.0}
        Mineral.__init__(self)


class nacf(Mineral):
    def __init__(self):
        formula = "NaAlSiO6"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "nacf",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -1965550.0,
            "S_0": 110.0,
            "V_0": 3.631e-05,
            "Cp": [272.7, -0.012398, 0.0, -2763.1],
            "a_0": 2.1e-05,
            "K_0": 1.85e11,
            "Kprime_0": 4.6,
            "Kdprime_0": -2.5e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 3440.0}
        Mineral.__init__(self)


class cacf(Mineral):
    def __init__(self):
        formula = "CaAl2O4"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "cacf",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -2325600.0,
            "S_0": 87.6,
            "V_0": 3.976e-05,
            "Cp": [191.9, 0.009563, -3211300.0, -640.2],
            "a_0": 1.93e-05,
            "K_0": 1.9e11,
            "Kprime_0": 4.0,
            "Kdprime_0": -2.1e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 10240.0}
        Mineral.__init__(self)


class manal(Mineral):
    def __init__(self):
        formula = "Mg3Al6O12"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "manal",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -6796630.0,
            "S_0": 250.0,
            "V_0": 0.00011166,
            "Cp": [600.0, 0.018756, -8989200.0, -2665.2],
            "a_0": 1.93e-05,
            "K_0": 1.84e11,
            "Kprime_0": 4.0,
            "Kdprime_0": -2.2e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 5120.0}
        Mineral.__init__(self)


class nanal(Mineral):
    def __init__(self):
        formula = "NaMg2SiAl5O12"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "nanal",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -6610270.0,
            "S_0": 280.0,
            "V_0": 0.00011322,
            "Cp": [672.7, 0.000106, -5992800.0, -4539.9],
            "a_0": 2.01e-05,
            "K_0": 1.84e11,
            "Kprime_0": 4.0,
            "Kdprime_0": -2.2e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 5120.0}
        Mineral.__init__(self)


class msnal(Mineral):
    def __init__(self):
        formula = "MgMg2Si3Mg3O12"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "msnal",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -6172380.0,
            "S_0": 272.5,
            "V_0": 0.00011061,
            "Cp": [639.9, 0.00807, -4231200.0, -4487.7],
            "a_0": 2.1e-05,
            "K_0": 1.85e11,
            "Kprime_0": 4.0,
            "Kdprime_0": -2.2e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 5120.0}
        Mineral.__init__(self)


class fsnal(Mineral):
    def __init__(self):
        formula = "FeFe2Si3Fe3O12"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "fsnal",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -4146000.0,
            "S_0": 440.2,
            "V_0": 0.00011856,
            "Cp": [543.3, 0.055578, -8301600.0, -1581.3],
            "a_0": 2.1e-05,
            "K_0": 1.85e11,
            "Kprime_0": 4.0,
            "Kdprime_0": -2.2e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 10240.0}
        Mineral.__init__(self)


class canal(Mineral):
    def __init__(self):
        formula = "CaMg2Al6O12"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "canal",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -6840000.0,
            "S_0": 257.6,
            "V_0": 0.00011159,
            "Cp": [591.9, 0.022067, -9204100.0, -2417.0],
            "a_0": 1.93e-05,
            "K_0": 1.77e11,
            "Kprime_0": 4.0,
            "Kdprime_0": -2.2e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 5120.0}
        Mineral.__init__(self)


class per(Mineral):
    def __init__(self):
        formula = "MgO"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "per",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -601570.0,
            "S_0": 26.5,
            "V_0": 1.125e-05,
            "Cp": [60.5, 0.000362, -535800.0, -299.2],
            "a_0": 3.11e-05,
            "K_0": 1.616e11,
            "Kprime_0": 3.95,
            "Kdprime_0": -2.4e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 260.0}
        Mineral.__init__(self)


class fper(Mineral):
    def __init__(self):
        formula = "FeO"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "fper",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -262240.0,
            "S_0": 58.6,
            "V_0": 1.206e-05,
            "Cp": [44.4, 0.00828, -1214200.0, 185.2],
            "a_0": 3.22e-05,
            "K_0": 1.52e11,
            "Kprime_0": 4.9,
            "Kdprime_0": -3.2e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 680.0}
        Mineral.__init__(self)


class cor(Mineral):
    def __init__(self):
        formula = "Al2O3"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "cor",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -1675250.0,
            "S_0": 50.9,
            "V_0": 2.558e-05,
            "Cp": [139.5, 0.00589, -2460600.0, -589.2],
            "a_0": 1.8e-05,
            "K_0": 2.54e11,
            "Kprime_0": 4.34,
            "Kdprime_0": -1.7e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 700.0}
        Mineral.__init__(self)


class mcor(Mineral):
    def __init__(self):
        formula = "MgSiO3"
        formula = dictionarize_formula(formula)
        self.params = {
            "name": "mcor",
            "formula": formula,
            "equation_of_state": "hp_tmt",
            "H_0": -1468000.0,
            "S_0": 59.3,
            "V_0": 2.635e-05,
            "Cp": [147.8, 0.002015, -2395000.0, -801.8],
            "a_0": 2.12e-05,
            "K_0": 2.11e11,
            "Kprime_0": 4.55,
            "Kdprime_0": -2.2e-11,
            "n": sum(formula.values()),
            "molar_mass": formula_mass(formula),
        }

        self.uncertainties = {"err_H_0": 880.0}
        Mineral.__init__(self)



from __future__ import print_function

# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

import numpy as np
from scipy.integrate import odeint
from scipy.integrate import quad
from scipy.interpolate import UnivariateSpline, interp1d
from scipy.optimize import fsolve
from burnman import constants
from burnman.utils import geotherm
import warnings

from .material import Material, material_property


class Layer(object):
    """
    The base class for a planetary layer.
    The user needs to set the following before properties can be computed:

    - set_material(), which sets the material of the layer,
      e.g. a mineral, solid_solution, or composite
    - set_temperature_mode(), either predefine, or set to an adiabatic profile
    - set_pressure_mode(), to set the self-consistent pressure
      (with user-defined option the pressures can be overwritten).
      To set the self-consistent pressure the pressure at the top and the
      gravity at the bottom of the layer need to be set.
    - make(), computes the self-consistent part of the layer and starts the
      settings to compute properties within the layer

    Note that the entire planet this layer sits in is not necessarily
    self-consistent, as the pressure at the top of the layer is a
    function of the density within the layer (through the gravity).
    Entire planets can be computed self-consistently with the planet class.
    Properties will be returned at the pre-defined radius array,
    although the evaluate() function can take a newly defined depthlist
    and values are interpolated between these (sufficient sampling of the layer
    is needed for this to be accurate).
    """

    def __init__(self, name=None, radii=None, verbose=False):
        self.name = name
        assert np.all(np.diff(radii) > 0)
        self.radii = radii
        self.outer_radius = max(self.radii)
        self.inner_radius = min(self.radii)
        self.thickness = self.outer_radius - self.inner_radius
        self.n_slices = len(self.radii)
        self.verbose = verbose
        self._cached = {}
        self._pressures = None
        self._temperatures = None
        self.sublayers = None
        self.material = None
        self.pressure_mode = "self-consistent"
        self.temperature_mode = None

    def __str__(self):
        """
        Prints details of the layer
        """
        writing = (
            f"The {self.name} is made of {self.material.name}"
            f" with {self.temperature_mode} temperatures and "
            f"{self.pressure_mode} pressures\n"
        )
        return writing

    def reset(self):
        """
        Resets all cached material properties.
        It is typically not required for the user to call this function.
        """
        self._cached = {}
        self._pressures = None
        self._temperatures = None
        self.sublayers = None

    def set_material(self, material):
        """
        Set the material of a Layer with a Material
        """
        assert isinstance(material, Material)
        self.material = material
        self.reset()

    def set_temperature_mode(
        self, temperature_mode="adiabatic", temperatures=None, temperature_top=None
    ):
        """
        Sets temperatures within the layer as user-defined values or as
        a (potentially perturbed) adiabat.

        :param temperature_mode: This can be set to 'user-defined', 'adiabatic',
            or 'perturbed-adiabatic'. 'user-defined' fixes the temperature
            with the profile input by the user. 'adiabatic' self-consistently
            computes the adiabat when setting the state of the layer.
            'perturbed-adiabatic' adds the user input array to the adiabat.
            This allows the user to apply boundary layers (for example).
        :type temperature_mode: string
        :param temperatures: The desired fixed temperatures in [K].
            Should have same length as defined radii in layer.
        :type temperatures: array of float
        :param temperature_top: Temperature at the top of the layer.
            Used if the temperature mode is chosen to be 'adiabatic' or
            'perturbed-adiabatic'. If 'perturbed-adiabatic' is chosen as the
            temperature mode, temperature_top corresponds to the true temperature
            at the top of the layer, and the reference isentrope at this radius
            is defined to lie at a temperature of
            temperature_top - temperatures[-1].
        :type temperature_top: float
        """
        self.reset()
        assert (
            temperature_mode == "user-defined"
            or temperature_mode == "adiabatic"
            or temperature_mode == "perturbed-adiabatic"
        )

        self.temperature_mode = temperature_mode

        if (
            temperature_mode == "user-defined"
            or temperature_mode == "perturbed-adiabatic"
        ):
            assert len(temperatures) == len(self.radii)
            self.usertemperatures = temperatures
        else:
            self.usertemperatures = np.zeros_like(self.radii)

        if temperature_mode == "adiabatic" or temperature_mode == "perturbed-adiabatic":
            self.temperature_top = temperature_top
        else:
            self.temperature_top = None

    def set_pressure_mode(
        self,
        pressure_mode="self-consistent",
        pressures=None,
        gravity_bottom=None,
        pressure_top=None,
        n_max_iterations=50,
        max_delta=1.0e-5,
    ):
        """
        Sets the pressure mode of the layer,
        which can either be 'user-defined', or 'self-consistent'.

        :param pressure_mode: This can be set to 'user-defined' or 'self-consistent'.
            'user-defined' fixes the pressures with the profile input
            by the user in the 'pressures' argument.
            'self-consistent' forces Layer to calculate pressures
            self-consistently. If this is selected, the user will need
            to supply values for the gravity_bottom [m/s^2]
            and pressure_top [Pa] arguments.
        :type pressure_mode: string
        :param pressures: Pressures [Pa] to set layer to
            (if the 'user-defined' pressure_mode has been selected).
            The array should be the same length as
            the layers user-defined radii array.
        :type pressures: array of floats
        :param pressure_top: Pressure [Pa] at the top of the layer.
        :type pressure_top: float
        :param gravity_bottom: Gravity [m/s^2] at the bottom of the layer.
        :type gravity_bottom: float
        :param n_max_iterations: Maximum number of iterations to reach
            self-consistent pressures.
        :type n_max_iterations: integer
        :param max_delta: Relative update to the highest pressure in the layer between
            iterations to stop iterations.
        :type max_delta: float
        """
        self.reset()
        assert pressure_mode == "user-defined" or pressure_mode == "self-consistent"
        self.pressure_mode = pressure_mode

        assert gravity_bottom is not None
        self.gravity_bottom = gravity_bottom

        if pressure_mode == "user-defined":
            assert pressures is not None
            assert len(pressures) == len(self.radii)
            self.pressures = pressures
            warnings.warn(
                "By setting the pressures in Layer they "
                "are unlikely to be self-consistent"
            )
        elif pressure_mode == "self-consistent":
            self.pressure_top = pressure_top
            self.n_max_iterations = n_max_iterations
            self.max_delta = max_delta
        else:
            raise NotImplementedError(
                f"pressure mode {pressure_mode} " "not recognised"
            )

    def make(self):
        """
        This routine needs to be called before evaluating any properties.
        If pressures and temperatures are not user-defined, they
        are computed here. This method also initializes an array of copied
        materials from which properties can be computed.
        """
        self.reset()
        if not hasattr(self, "material"):
            raise AttributeError(
                "You must set_material() for the layer " "before running make()."
            )
        if not hasattr(self, "temperature_mode"):
            raise AttributeError(
                "You must set_temperature_mode() for the "
                "layer before running make()."
            )
        if not hasattr(self, "pressure_mode"):
            raise AttributeError(
                "You must set_pressure_mode() for the layer " "before running make()."
            )

        if self.pressure_mode == "user-defined":
            self.temperatures = self._evaluate_temperature(
                self.pressures, self.temperature_top
            )
        elif self.pressure_mode == "self-consistent":
            new_press = (
                self.pressure_top + (-self.radii + max(self.radii)) * 1.0e3
            )  # initial pressure curve guess
            temperatures = self._evaluate_temperature(new_press, self.temperature_top)
            # Make it self-consistent!!!
            i = 0

            while i < self.n_max_iterations:
                i += 1
                ref_press = new_press
                new_grav, new_press = self._evaluate_eos(
                    new_press, temperatures, self.gravity_bottom, self.pressure_top
                )
                temperatures = self._evaluate_temperature(
                    new_press, self.temperature_top
                )
                rel_err = abs((max(ref_press) - max(new_press)) / max(new_press))
                if self.verbose:
                    print(
                        f"Iteration {i:0d} maximum relative pressure error: "
                        f"{rel_err:.1f}"
                    )

                if rel_err < self.max_delta:
                    break

            self.pressures = new_press
            self.temperatures = temperatures
        else:
            raise NotImplementedError("pressure mode not recognised")

        self.sublayers = []
        for r in range(len(self.radii)):
            self.sublayers.append(self.material.copy())
            self.sublayers[r].set_state(self.pressures[r], self.temperatures[r])

    def evaluate(self, properties, radlist=None, radius_planet=None):
        """
        Function that is used to evaluate properties
        across the layer. If radlist is not defined, values are
        returned at the internal radlist.
        If asking for different radii than the internal radlist,
        pressure and temperature values are interpolated and the
        layer material evaluated at those pressures and
        temperatures.

        :param properties: List of properties to evaluate.
        :type properties: list of strings
        :param radlist: Radii to evaluate properties at. If left empty,
            internal radii list is used.
        :type radlist: array of floats
        :param planet_radius: Planet outer radius. Used only to calculate depth.
        :type planet_radius: float

        :returns: 1D or 2D array of requested properties
            (1D if only one property was requested)
        :rtype: numpy.array
        """

        if radlist is None:
            values = np.empty([len(properties), len(self.radii)])
            for i, prop in enumerate(properties):
                if prop == "depth":
                    values[i] = radius_planet - self.radii
                else:
                    try:
                        values[i] = getattr(self, prop)
                    except:
                        values[i] = np.array(
                            [
                                getattr(self.sublayers[i], prop)
                                for i in range(len(self.sublayers))
                            ]
                        )
        else:
            func_p = interp1d(self.radii, self.pressures)
            pressures = func_p(radlist)
            func_t = interp1d(self.radii, self.temperatures)
            temperatures = func_t(radlist)
            values = np.empty([len(properties), len(radlist)])
            for i, prop in enumerate(properties):
                if prop == "depth":
                    values[i] = radius_planet - radlist
                else:
                    try:
                        values[i] = self.material.evaluate(
                            [prop], pressures, temperatures
                        )
                    except:
                        func_prop = interp1d(self.radii, getattr(self, prop))
                        values[i] = func_prop(radlist)

        if values.shape[0] == 1:
            values = values[0]
        return values

    def _evaluate_temperature(self, pressures=None, temperature_top=None):
        """
        Returns the temperatures of the layer for given pressures.
        Used by make()
        """
        if (
            self.temperature_mode == "adiabatic"
            or self.temperature_mode == "perturbed-adiabatic"
        ):
            adiabat = geotherm.adiabatic(
                pressures[::-1],
                temperature_top - self.usertemperatures[-1],
                self.material,
            )[::-1]
        else:
            adiabat = np.zeros_like(self.radii)
        return adiabat + self.usertemperatures

    def _evaluate_eos(self, pressures, temperatures, gravity_bottom, pressure_top):
        """
        Returns updated gravity and pressure
        make() loops over this until consistency is achieved.
        """
        [density] = self.material.evaluate(["density"], pressures, temperatures)
        grav = self._compute_gravity(density, gravity_bottom)
        press = self._compute_pressure(density, grav, pressure_top)
        return grav, press

    # Functions needed to compute self-consistent radii-pressures
    def _compute_gravity(self, density, gravity_bottom):
        """
        Computes the gravity of a layer
        Used by _evaluate_eos()
        """
        # Create a spline fit of density as a function of radius
        rhofunc = UnivariateSpline(self.radii, density)
        # Numerically integrate Poisson's equation

        def poisson(p, x):
            return 4.0 * np.pi * constants.G * rhofunc(x) * x * x

        grav = np.ravel(
            odeint(poisson, gravity_bottom * self.radii[0] * self.radii[0], self.radii)
        )

        if self.radii[0] == 0:
            grav[0] = 0
            grav[1:] = grav[1:] / self.radii[1:] / self.radii[1:]
        else:
            grav[:] = grav[:] / self.radii[:] / self.radii[:]
        return grav

    def _compute_pressure(self, density, gravity, pressure_top):
        """
        Calculate the pressure profile based on density and gravity.
        This integrates the equation for hydrostatic equilibrium P = rho g z.
        Used by _evaluate_eos()
        """
        # flip radius, density and gravity to increasing pressure
        depthfromtop = -self.radii[::-1] + max(self.radii)
        density = density[::-1]
        gravity = gravity[::-1]
        # Make a spline fit of density as a function of depth
        rhofunc = UnivariateSpline(depthfromtop, density)
        # Make a spline fit of gravity as a function of depth
        gfunc = UnivariateSpline(depthfromtop, gravity)

        # integrate the hydrostatic equation
        pressure = np.ravel(
            odeint((lambda p, x: gfunc(x) * rhofunc(x)), pressure_top, depthfromtop)
        )

        return pressure[::-1]

    @property
    def mass(self):
        """
        Calculates the mass of the layer [kg]
        """
        mass = 0.0
        radii = self.radii
        density = self.evaluate(["density"])
        rhofunc = UnivariateSpline(radii, density)
        mass = np.abs(
            quad(lambda r: 4 * np.pi * rhofunc(r) * r * r, radii[0], radii[-1])[0]
        )
        return mass

    @property
    def moment_of_inertia(self):
        """
        Returns the moment of inertia of the layer [kg m^2]
        """
        moment = 0.0
        radii = self.radii
        density = self.evaluate(["density"])
        rhofunc = UnivariateSpline(radii, density)
        moment = np.abs(
            quad(
                lambda r: 8.0 / 3.0 * np.pi * rhofunc(r) * r * r * r * r,
                radii[0],
                radii[-1],
            )[0]
        )
        return moment

    @property
    def gravity(self):
        """
        Returns gravity profile of the layer [m s^(-2)]
        """
        return self._compute_gravity(self.density, self.gravity_bottom)

    @property
    def bullen(self):
        """
        Returns the Bullen parameter across the layer.
        The Bullen parameter assess if compression as a function of pressure is
        like homogeneous, adiabatic compression.
        Bullen parameter =1  , homogeneous, adiabatic compression
        Bullen parameter > 1 , more compressed with pressure,
        e.g. across phase transitions Bullen parameter < 1,
        less compressed with pressure, e.g. across a boundary layer.
        """
        kappa = self.bulk_sound_velocity * self.bulk_sound_velocity * self.density
        phi = self.bulk_sound_velocity * self.bulk_sound_velocity
        try:
            dkappadP = np.gradient(kappa, edge_order=2) / np.gradient(
                self.pressures, edge_order=2
            )
            dphidr = (
                np.gradient(phi, edge_order=2)
                / np.gradient(self.radii, edge_order=2)
                / self.gravity
            )
        except:
            dkappadP = np.gradient(kappa) / np.gradient(self.pressures)
            dphidr = np.gradient(phi) / np.gradient(self.radii) / self.gravity
        bullen = dkappadP + dphidr
        return bullen

    @property
    def brunt_vasala(self):
        """
        Returns the brunt-vasala (or buoyancy) frequency, N, across the layer.
        This frequency assess the stabilty of the layer:
        N < 0, fluid will convect
        N= 0, fluid is neutral
        N > 0, fluid is stabily stratified.
        """
        kappa = self.bulk_sound_velocity * self.bulk_sound_velocity * self.density
        brunt_vasala = (
            self.density * self.gravity * self.gravity * (self.bullen - 1.0) / kappa
        )
        return brunt_vasala

    @property
    def pressure(self):
        """
        Returns current pressures across the layer that was set
        with :func:`~burnman.Material.set_state`.

        Aliased with :func:`~burnman.Material.P`.

        :returns: Pressures in [Pa] at the predefined radii.
        :rtype: numpy.array
        """
        return self.pressures

    @property
    def temperature(self):
        """
        Returns current temperature  across the layer that was set with
        :func:`~burnman.Material.set_state`.

        - Aliased with :func:`~burnman.Material.T`.

        :returns: Temperatures in [K] at the predefined radii.
        :rtype: numpy.array
        """
        return self.temperatures

    @material_property
    def molar_internal_energy(self):
        """
        Returns the molar internal energies across the layer.

        Notes
        -----
        - Needs to be implemented in derived classes.
        - Aliased with :func:`~burnman.Material.energy`.

        :returns: The internal energies in [J/mol] at the predefined radii.
        :rtype: numpy.array
        """
        return np.array(
            [
                self.sublayers[i].molar_internal_energy
                for i in range(len(self.sublayers))
            ]
        )

    @material_property
    def molar_gibbs(self):
        """
        Returns the molar Gibbs free energies across the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.gibbs`.

        :returns: Gibbs energies in [J/mol] at the predefined radii.
        :rtype: numpy.array
        """
        return np.array(
            [self.sublayers[i].molar_gibbs for i in range(len(self.sublayers))]
        )

    @material_property
    def molar_helmholtz(self):
        """
        Returns the molar Helmholtz free energies across the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.helmholtz`.

        :returns: Helmholtz energies in [J/mol] at the predefined radii.
        :rtype: numpy.array
        """
        return np.array(
            [self.sublayers[i].molar_helmholtz for i in range(len(self.sublayers))]
        )

    @material_property
    def molar_mass(self):
        """
        Returns molar mass of the layer.

        Needs to be implemented in derived classes.

        :returns: Molar mass in [kg/mol].
        :rtype: numpy.array
        """
        return np.array(
            [self.sublayers[i].molar_mass for i in range(len(self.sublayers))]
        )

    @material_property
    def molar_volume(self):
        """
        Returns molar volumes across the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.V`.

        :returns: Molar volumes in [m^3/mol] at the predefined radii.
        :rtype: numpy.array
        """
        return np.array(
            [self.sublayers[i].molar_volume for i in range(len(self.sublayers))]
        )

    @material_property
    def density(self):
        """
        Returns the densities across this layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.rho`.

        :returns: The densities of this material in [kg/m^3] at the predefined radii.
        :rtype: numpy.array
        """
        return np.array([self.sublayers[i].density for i in range(len(self.sublayers))])

    @material_property
    def molar_entropy(self):
        """
        Returns molar entropies acroos the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.S`.

        :returns: Entropies in [J/K/mol] at the predefined radii.
        :rtype: numpy.array
        """
        return np.array(
            [self.sublayers[i].molar_entropy for i in range(len(self.sublayers))]
        )

    @material_property
    def molar_enthalpy(self):
        """
        Returns molar enthalpies across the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.H`.

        :returns: Enthalpies in [J/mol] at the predefined radii.
        :rtype: numpy.array
        """
        return np.array(
            [self.sublayers[i].molar_enthalpy for i in range(len(self.sublayers))]
        )

    @material_property
    def isothermal_bulk_modulus_reuss(self):
        """
        Returns isothermal bulk moduli across the layer.

        Notes
        -----
        - Needs to be implemented in derived classes.
        - Aliased with :func:`~burnman.Material.K_T`.

        :returns: Bulk moduli in [Pa] at the predefined radii.
        :rtype: numpy.array
        """
        return np.array(
            [
                self.sublayers[i].isothermal_bulk_modulus_reuss
                for i in range(len(self.sublayers))
            ]
        )

    @material_property
    def isentropic_bulk_modulus_reuss(self):
        """
        Returns the adiabatic bulk moduli across the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.K_S`.

        :returns: Adiabatic bulk modulus in [Pa] at the predefined radii.
        :rtype: numpy.array
        """
        return np.array(
            [
                self.sublayers[i].isentropic_bulk_modulus_reuss
                for i in range(len(self.sublayers))
            ]
        )

    @material_property
    def isothermal_compressibility_reuss(self):
        """
        Returns isothermal compressibilities across the layer
        (or inverse isothermal bulk moduli).

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.beta_T`.

        :returns: Isothermal compressibilities in [1/Pa] at the predefined radii.
        :rtype: numpy.array
        """
        return np.array(
            [
                self.sublayers[i].isothermal_compressibility_reuss
                for i in range(len(self.sublayers))
            ]
        )

    @material_property
    def isentropic_compressibility_reuss(self):
        """
        Returns adiabatic compressibilities across the layer
        (or inverse adiabatic bulk moduli).

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.beta_S`.

        :returns: Adiabatic compressibilities in [1/Pa] at the predefined radii.
        :rtype: numpy.array
        """
        return np.array(
            [
                self.sublayers[i].isentropic_compressibility_reuss
                for i in range(len(self.sublayers))
            ]
        )

    @material_property
    def shear_modulus(self):
        """
        Returns shear moduli across the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.beta_G`.

        :returns: Shear moduli in [Pa] at the predefined radii.
        :rtype: numpy.array
        """
        return np.array(
            [self.sublayers[i].shear_modulus for i in range(len(self.sublayers))]
        )

    @material_property
    def p_wave_velocity(self):
        """
        Returns P wave speeds across the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.v_p`.

        :returns: P wave speeds in [m/s] at the predefined radii.
        :rtype: numpy.array
        """
        return np.array(
            [self.sublayers[i].p_wave_velocity for i in range(len(self.sublayers))]
        )

    @material_property
    def bulk_sound_velocity(self):
        """
        Returns bulk sound speeds across the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.v_phi`.

        :returns: Bulk sound velocities in [m/s] at the predefined radii.
        :rtype: numpy.array
        """
        return np.array(
            [self.sublayers[i].bulk_sound_velocity for i in range(len(self.sublayers))]
        )

    @material_property
    def shear_wave_velocity(self):
        """
        Returns shear wave speeds across the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.v_s`.

        :returns: Shear wave speeds in [m/s] at the predefined radii.
        :rtype: numpy.array
        """
        return np.array(
            [self.sublayers[i].shear_wave_velocity for i in range(len(self.sublayers))]
        )

    @material_property
    def grueneisen_parameter(self):
        """
        Returns the grueneisen parameters across the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.gr`.

        :returns: Grueneisen parameters [unitless] at the predefined radii.
        :rtype: numpy.array
        """
        return np.array(
            [self.sublayers[i].grueneisen_parameter for i in range(len(self.sublayers))]
        )

    @material_property
    def thermal_expansivity(self):
        """
        Returns thermal expansion coefficients across the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.alpha`.

        :returns: Thermal expansivities in [1/K] at the predefined radii.
        :rtype: numpy.array
        """
        return np.array(
            [self.sublayers[i].thermal_expansivity for i in range(len(self.sublayers))]
        )

    @material_property
    def molar_heat_capacity_v(self):
        """
        Returns molar heat capacity at constant volumes across the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.C_v`.

        :returns: Heat capacities in [J/K/mol] at the predefined radii.
        :rtype: numpy.array
        """
        return np.array(
            [
                self.sublayers[i].molar_heat_capacity_v
                for i in range(len(self.sublayers))
            ]
        )

    @material_property
    def molar_heat_capacity_p(self):
        """
        Returns molar_heat capacity at constant pressures across the layer.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.C_p`.

        :returns: Heat capacities in [J/K/mol] at the predefined radii.
        :rtype: numpy.array
        """
        return np.array(
            [
                self.sublayers[i].molar_heat_capacity_p
                for i in range(len(self.sublayers))
            ]
        )

    # Aliased properties
    @property
    def P(self):
        """Alias for :func:`~burnman.Layer.pressure`"""
        return self.pressure

    @property
    def T(self):
        """Alias for :func:`~burnman.Layer.temperature`"""
        return self.temperature

    @property
    def energy(self):
        """Alias for :func:`~burnman.Layer.molar_internal_energy`"""
        return self.molar_internal_energy

    @property
    def helmholtz(self):
        """Alias for :func:`~burnman.Layer.molar_helmholtz`"""
        return self.molar_helmholtz

    @property
    def gibbs(self):
        """Alias for :func:`~burnman.Layer.molar_gibbs`"""
        return self.molar_gibbs

    @property
    def V(self):
        """Alias for :func:`~burnman.Layer.molar_volume`"""
        return self.molar_volume

    @property
    def rho(self):
        """Alias for :func:`~burnman.Layer.density`"""
        return self.density

    @property
    def S(self):
        """Alias for :func:`~burnman.Layer.molar_entropy`"""
        return self.molar_entropy

    @property
    def H(self):
        """Alias for :func:`~burnman.Layer.molar_enthalpy`"""
        return self.molar_enthalpy

    @property
    def K_T(self):
        """Alias for :func:`~burnman.Layer.isothermal_bulk_modulus_reuss`"""
        return self.isothermal_bulk_modulus_reuss

    @property
    def K_S(self):
        """Alias for :func:`~burnman.Layer.isentropic_bulk_modulus_reuss`"""
        return self.isentropic_bulk_modulus_reuss

    @property
    def beta_T(self):
        """Alias for :func:`~burnman.Layer.isothermal_compressibility_reuss`"""
        return self.isothermal_compressibility_reuss

    @property
    def beta_S(self):
        """Alias for :func:`~burnman.Layer.isentropic_compressibility_reuss`"""
        return self.isentropic_compressibility_reuss

    @property
    def G(self):
        """Alias for :func:`~burnman.Layer.shear_modulus`"""
        return self.shear_modulus

    @property
    def v_p(self):
        """Alias for :func:`~burnman.Layer.p_wave_velocity`"""
        return self.p_wave_velocity

    @property
    def v_phi(self):
        """Alias for :func:`~burnman.Layer.bulk_sound_velocity`"""
        return self.bulk_sound_velocity

    @property
    def v_s(self):
        """Alias for :func:`~burnman.Layer.shear_wave_velocity`"""
        return self.shear_wave_velocity

    @property
    def gr(self):
        """Alias for :func:`~burnman.Layer.grueneisen_parameter`"""
        return self.grueneisen_parameter

    @property
    def alpha(self):
        """Alias for :func:`~burnman.Layer.thermal_expansivity`"""
        return self.thermal_expansivity

    @property
    def C_v(self):
        """Alias for :func:`~burnman.Material.molar_heat_capacity_v`"""
        return self.molar_heat_capacity_v

    @property
    def C_p(self):
        """Alias for :func:`~burnman.Material.molar_heat_capacity_p`"""
        return self.molar_heat_capacity_p


class BoundaryLayerPerturbation(object):
    """
    A class that implements a temperature perturbation model corresponding to a
    simple thermal boundary layer.
    The model takes the following form:
    T = a*exp((r - r1)/(r0 - r1)*c) + b*exp((r - r0)/(r1 - r0)*c)
    The relationships between the input parameters and a, b and c are
    given below.

    This model is a simpler version of that proposed
    by :cite:`Richter1981`.

    :param radius_bottom: The radius at the bottom of the layer (r0) [m].
    :type radius_bottom: float

    :param radius_top: The radius at the top of the layer (r1) [m].
    :type radius_top: float

    :param rayleigh_number: The Rayleigh number of convection within the layer. The
        exponential scale factor is this number to the power of 1/4
        (Ra = c^4).
    :type rayleigh_number: float

    :param temperature_change: The total difference in potential
        temperature across the layer [K]. temperature_change = (a + b)*exp(c).
    :type temperature_change: float

    :param boundary_layer_ratio: The ratio of the linear scale factors (a/b)
        corresponding to the thermal boundary layers at the top and bottom of
        the layer. A number greater than 1 implies a larger change in
        temperature across the top boundary than the bottom boundary.
    :type boundary_layer_ratio: float
    """

    def __init__(
        self,
        radius_bottom,
        radius_top,
        rayleigh_number,
        temperature_change,
        boundary_layer_ratio,
    ):
        self.r0 = radius_bottom
        self.r1 = radius_top

        self.Ra = rayleigh_number
        self.c = np.power(self.Ra, 1.0 / 4.0)

        self.a = temperature_change / (np.exp(self.c) * (1.0 + boundary_layer_ratio))
        self.b = -boundary_layer_ratio * self.a

    def temperature(self, radii):
        """
        Returns the temperature at one or more radii [K].

        :param radii: The radii at which to evaluate the temperature.
        :type radii: float or numpy.array

        :returns: The temperatures at the requested radii.
        :rtype: float or numpy.array
        """
        return self.a * np.exp(
            (radii - self.r1) / (self.r0 - self.r1) * self.c
        ) + self.b * np.exp((radii - self.r0) / (self.r1 - self.r0) * self.c)

    def dTdr(self, radii):
        """
        Returns the thermal gradient at one or more radii [K/m].

        :param radii: The radii at which to evaluate the thermal gradients.
        :type radii: float or numpy.array

        :returns: The thermal gradient at the requested radii.
        :rtype: float or numpy.array
        """
        return (
            self.c
            / (self.r0 - self.r1)
            * (
                self.a * np.exp((radii - self.r1) / (self.r0 - self.r1) * self.c)
                - self.b * np.exp((radii - self.r0) / (self.r1 - self.r0) * self.c)
            )
        )

    def set_model_thermal_gradients(self, dTdr_bottom, dTdr_top):
        """
        Reparameterizes the model based on the thermal gradients
        at the bottom and top of the model.

        :param dTdr_bottom: The thermal gradient at the bottom of the model [K/m].
            Typically negative for a cooling planet.
        :type dTdr_bottom: float

        :param dTdr_top: The thermal gradient at the top of the model [K/m].
            Typically negative for a cooling planet.
        :type dTdr_top: float
        """

        def delta_dTdrs(args, dTdr_bottom, dTdr_top):
            a, b = args
            self.a = a
            self.b = b

            return [dTdr_bottom - self.dTdr(self.r0), dTdr_top - self.dTdr(self.r1)]

        fsolve(delta_dTdrs, [self.a, self.b], args=(dTdr_bottom, dTdr_top))
